@@ -37,19 +37,39 @@
  *		    is a signed byte rather than a word.  Opcode byte follows.
  * \150..\153     - an immediate dword or signed byte for operand 0..3
  * \154..\157    - or 2 (s-field) into opcode byte if operand 0..3
- *		    is a signed byte rather than a word.  Opcode byte follows.
+ *		    is a signed byte rather than a dword.  Opcode byte follows.
  * \160..\163    - this instruction uses DREX rather than REX, with the
  *		   OC0 field set to 0, and the dest field taken from
  *                 operand 0..3.
  * \164..\167    - this instruction uses DREX rather than REX, with the
  *		   OC0 field set to 1, and the dest field taken from
  *                 operand 0..3.
- * \170          - encodes the literal byte 0. (Some compilers don't take
- *                 kindly to a zero byte in the _middle_ of a compile time
- *                 string constant, so I had to put this hack in.)
  * \171		 - placement of DREX suffix in the absence of an EA
+ * \172\ab	 - the register number from operand a in bits 7..4, with
+ *                 the 4-bit immediate from operand b in bits 3..0.
+ * \173\xab	 - the register number from operand a in bits 7..4, with
+ *		   the value b in bits 3..0.
+ * \174\a	 - the register number from operand a in bits 7..4, and
+ *		   an arbitrary value in bits 3..0 (assembled as zero.)
  * \2ab          - a ModRM, calculated on EA in operand a, with the spare
  *                 field equal to digit b.
+ * \250..\253    - same as \150..\153, except warn if the 64-bit operand
+ *                 is not equal to the truncated and sign-extended 32-bit
+ *                 operand; used for 32-bit immediates in 64-bit mode.
+ * \260..\263    - this instruction uses VEX rather than REX, with the
+ *		   V field taken from operand 0..3.
+ * \270		 - this instruction uses VEX rather than REX, with the
+ *		   V field set to 1111b.
+ *
+ * VEX prefixes are followed by the sequence:
+ * \mm\wlp         where mm is the M field; and wlp is:
+ *                 00 0ww lpp
+ *                 [w0] ww = 0 for W = 0
+ *                 [w1] ww = 1 for W = 1
+ *                 [wx] ww = 2 for W don't care (always assembled as 0)
+ *                 [ww] ww = 3 for W used as REX.W
+ *
+ *
  * \310          - indicates fixed 16-bit address size, i.e. optional 0x67.
  * \311          - indicates fixed 32-bit address size, i.e. optional 0x67.
  * \312          - (disassembler only) marker on LOOP, LOOPxx instructions.
@@ -75,6 +95,10 @@
  * \335          - disassemble a rep (0xF3 byte) prefix as repe not rep.
  * \340          - reserve <operand 0> bytes of uninitialized storage.
  *                 Operand 0 had better be a segmentless constant.
+ * \360		 - no SSE prefix (== \364\331)
+ * \361          - 66 SSE prefix (== \366\331)
+ * \362          - F2 SSE prefix (== \364\332)
+ * \363          - F3 SSE prefix (== \364\333)
  * \364          - operand-size prefix (0x66) not permitted
  * \365          - address-size prefix (0x67) not permitted
  * \366          - operand-size prefix (0x66) used as opcode extension
@@ -95,9 +119,10 @@
 #include "nasmlib.h"
 #include "assemble.h"
 #include "insns.h"
-#include "preproc.h"
-#include "regflags.c"
-#include "regvals.c"
+#include "tables.h"
+
+/* Initialized to zero by the C standard */
+static const uint8_t const_zero_buf[256];
 
 typedef struct {
     int sib_present;                 /* is a SIB byte necessary? */
@@ -111,8 +136,8 @@ static efunc errfunc;
 static struct ofmt *outfmt;
 static ListGen *list;
 
-static int64_t calcsize(int32_t, int64_t, int, insn *, const char *);
-static void gencode(int32_t, int64_t, int, insn *, const char *, int64_t);
+static int64_t calcsize(int32_t, int64_t, int, insn *, const uint8_t *);
+static void gencode(int32_t, int64_t, int, insn *, const uint8_t *, int64_t);
 static int matches(const struct itemplate *, insn *, int bits);
 static int32_t regflag(const operand *);
 static int32_t regval(const operand *);
@@ -148,6 +173,8 @@ static const char *size_name(int size)
 	return "tword";
     case 16:
 	return "oword";
+    case 32:
+	return "yword";
     default:
 	return "???";
     }
@@ -159,7 +186,8 @@ static void warn_overflow(int size, int64_t data)
 	int64_t lim = ((int64_t)1 << (size*8))-1;
 
 	if (data < ~lim || data > lim)
-	    errfunc(ERR_WARNING | ERR_WARN_NOV, "%s data exceeds bounds", size_name(size));
+	    errfunc(ERR_WARNING | ERR_WARN_NOV,
+		    "%s data exceeds bounds", size_name(size));
     }
 }
 /*
@@ -186,7 +214,7 @@ static void out(int64_t offset, int32_t segto, const void *data,
 	    errfunc(ERR_PANIC, "OUT_ADDRESS with size > 8");
 	    return;
 	}
-	
+
 	WRITEADDR(q, *(int64_t *)data, size);
 	data = p;
 	type = OUT_RAWDATA;
@@ -211,7 +239,7 @@ static void out(int64_t offset, int32_t segto, const void *data,
 }
 
 static int jmp_match(int32_t segment, int64_t offset, int bits,
-                     insn * ins, const char *code)
+                     insn * ins, const uint8_t *code)
 {
     int64_t isize;
     uint8_t c = code[0];
@@ -273,6 +301,9 @@ int64_t assemble(int32_t segment, int64_t offset, int bits, uint32_t cp,
     case I_DO:
 	wsize = 16;
 	break;
+    case I_DY:
+	wsize = 32;
+	break;
     default:
 	break;
     }
@@ -297,7 +328,8 @@ int64_t assemble(int32_t segment, int64_t offset, int bits, uint32_t cp,
                                 OUT_RAWDATA, 1, NO_SEG, NO_SEG);
                         }
                     } else if (wsize > 8) {
-                        errfunc(ERR_NONFATAL, "integer supplied to a DT or DO"
+                        errfunc(ERR_NONFATAL,
+				"integer supplied to a DT, DO or DY"
                                 " instruction");
                     } else
                         out(offset, segment, &e->offset,
@@ -312,8 +344,7 @@ int64_t assemble(int32_t segment, int64_t offset, int bits, uint32_t cp,
 
                     if (align) {
                         align = wsize - align;
-                        out(offset, segment,
-			    "\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0",
+                        out(offset, segment, const_zero_buf,
                             OUT_RAWDATA, align, NO_SEG, NO_SEG);
                     }
                     offset += e->stringlen + align;
@@ -337,8 +368,6 @@ int64_t assemble(int32_t segment, int64_t offset, int bits, uint32_t cp,
         static char fname[FILENAME_MAX];
         FILE *fp;
         int32_t len;
-        char *prefix = "", *combine;
-        char **pPrevPath = NULL;
 
         len = FILENAME_MAX - 1;
         if (len > instruction->eops->stringlen)
@@ -346,30 +375,14 @@ int64_t assemble(int32_t segment, int64_t offset, int bits, uint32_t cp,
         strncpy(fname, instruction->eops->stringval, len);
         fname[len] = '\0';
 
-        while (1) {         /* added by alexfru: 'incbin' uses include paths */
-            combine = nasm_malloc(strlen(prefix) + len + 1);
-            strcpy(combine, prefix);
-            strcat(combine, fname);
-
-            if ((fp = fopen(combine, "rb")) != NULL) {
-                nasm_free(combine);
-                break;
-            }
-
-            nasm_free(combine);
-            pPrevPath = pp_get_include_path_ptr(pPrevPath);
-            if (pPrevPath == NULL)
-                break;
-            prefix = *pPrevPath;
-        }
-
-        if (fp == NULL)
+	fp = fopen(fname, "rb");
+	if (!fp) {
             error(ERR_NONFATAL, "`incbin': unable to open file `%s'",
                   fname);
-        else if (fseek(fp, 0L, SEEK_END) < 0)
+	} else if (fseek(fp, 0L, SEEK_END) < 0) {
             error(ERR_NONFATAL, "`incbin': unable to seek on file `%s'",
                   fname);
-        else {
+	} else {
             static char buf[2048];
             int32_t t = instruction->times;
             int32_t base = 0;
@@ -442,7 +455,7 @@ int64_t assemble(int32_t segment, int64_t offset, int bits, uint32_t cp,
             m += jmp_match(segment, offset, bits, instruction, temp->code);
 
         if (m == 100) {         /* matches! */
-            const char *codes = temp->code;
+            const uint8_t *codes = temp->code;
             int64_t insn_size = calcsize(segment, offset, bits,
                                       instruction, codes);
             itimes = instruction->times;
@@ -609,7 +622,8 @@ int64_t insn_size(int32_t segment, int64_t offset, int bits, uint32_t cp,
 
     if (instruction->opcode == I_DB || instruction->opcode == I_DW ||
         instruction->opcode == I_DD || instruction->opcode == I_DQ ||
-	instruction->opcode == I_DT || instruction->opcode == I_DO) {
+	instruction->opcode == I_DT || instruction->opcode == I_DO ||
+	instruction->opcode == I_DY) {
         extop *e;
         int32_t isize, osize, wsize = 0;   /* placate gcc */
 
@@ -633,6 +647,9 @@ int64_t insn_size(int32_t segment, int64_t offset, int bits, uint32_t cp,
 	case I_DO:
 	    wsize = 16;
 	    break;
+	case I_DY:
+	    wsize = 32;
+	    break;
 	default:
 	    break;
         }
@@ -658,34 +675,15 @@ int64_t insn_size(int32_t segment, int64_t offset, int bits, uint32_t cp,
         char fname[FILENAME_MAX];
         FILE *fp;
         int32_t len;
-        char *prefix = "", *combine;
-        char **pPrevPath = NULL;
 
         len = FILENAME_MAX - 1;
         if (len > instruction->eops->stringlen)
             len = instruction->eops->stringlen;
         strncpy(fname, instruction->eops->stringval, len);
         fname[len] = '\0';
-
-	/* added by alexfru: 'incbin' uses include paths */
-        while (1) {
-            combine = nasm_malloc(strlen(prefix) + len + 1);
-            strcpy(combine, prefix);
-            strcat(combine, fname);
-
-            if ((fp = fopen(combine, "rb")) != NULL) {
-                nasm_free(combine);
-                break;
-            }
-
-            nasm_free(combine);
-            pPrevPath = pp_get_include_path_ptr(pPrevPath);
-            if (pPrevPath == NULL)
-                break;
-            prefix = *pPrevPath;
-        }
-
-        if (fp == NULL)
+	
+	fp = fopen(fname, "rb");
+	if (!fp)
             error(ERR_NONFATAL, "`incbin': unable to open file `%s'",
                   fname);
         else if (fseek(fp, 0L, SEEK_END) < 0)
@@ -717,7 +715,7 @@ int64_t insn_size(int32_t segment, int64_t offset, int bits, uint32_t cp,
         if (m == 100) {
             /* we've matched an instruction. */
             int64_t isize;
-            const char *codes = temp->code;
+            const uint8_t *codes = temp->code;
             int j;
 
             isize = calcsize(segment, offset, bits, instruction, codes);
@@ -756,27 +754,57 @@ int64_t insn_size(int32_t segment, int64_t offset, int bits, uint32_t cp,
     return -1;                  /* didn't match any instruction */
 }
 
-/* check that  opn[op]  is a signed byte of size 16 or 32,
-					and return the signed value*/
-static int is_sbyte(insn * ins, int op, int size)
+static bool possible_sbyte(insn * ins, int op)
 {
-    int32_t v;
-    int ret;
-
-    ret = !(ins->forw_ref && ins->oprs[op].opflags) &&  /* dead in the water on forward reference or External */
+    return !(ins->forw_ref && ins->oprs[op].opflags) &&
         optimizing >= 0 &&
         !(ins->oprs[op].type & STRICT) &&
         ins->oprs[op].wrt == NO_SEG && ins->oprs[op].segment == NO_SEG;
-
-    v = ins->oprs[op].offset;
-    if (size == 16)
-        v = (int16_t)v;    /* sign extend if 16 bits */
-
-    return ret && v >= -128L && v <= 127L;
 }
 
+/* check that opn[op]  is a signed byte of size 16 or 32 */
+static bool is_sbyte16(insn * ins, int op)
+{
+    int16_t v;
+
+    if (!possible_sbyte(ins, op))
+	return false;
+
+    v = ins->oprs[op].offset;
+    return v >= -128 && v <= 127;
+}
+
+static bool is_sbyte32(insn * ins, int op)
+{
+    int32_t v;
+
+    if (!possible_sbyte(ins, op))
+	return false;
+
+    v = ins->oprs[op].offset;
+    return v >= -128 && v <= 127;
+}
+
+/* check that  opn[op]  is a signed byte of size 32; warn if this is not
+   the original value when extended to 64 bits */
+static bool is_sbyte64(insn * ins, int op)
+{
+    int64_t v64;
+    int32_t v32;
+
+    /* dead in the water on forward reference or External */
+    if (!possible_sbyte(ins, op))
+	return false;
+
+    v64 = ins->oprs[op].offset;
+    v32 = (int32_t)v64;
+
+    warn_overflow(32, v64);
+
+    return v32 >= -128 && v32 <= 127;
+}
 static int64_t calcsize(int32_t segment, int64_t offset, int bits,
-                     insn * ins, const char *codes)
+                     insn * ins, const uint8_t *codes)
 {
     int64_t length = 0;
     uint8_t c;
@@ -902,7 +930,7 @@ static int64_t calcsize(int32_t segment, int64_t offset, int bits,
         case 0141:
         case 0142:
 	case 0143:
-            length += is_sbyte(ins, c & 3, 16) ? 1 : 2;
+            length += is_sbyte16(ins, c & 3) ? 1 : 2;
             break;
         case 0144:
         case 0145:
@@ -915,7 +943,7 @@ static int64_t calcsize(int32_t segment, int64_t offset, int bits,
         case 0151:
         case 0152:
         case 0153:
-            length += is_sbyte(ins, c & 3, 32) ? 1 : 4;
+            length += is_sbyte32(ins, c & 3) ? 1 : 4;
             break;
         case 0154:
         case 0155:
@@ -930,7 +958,7 @@ static int64_t calcsize(int32_t segment, int64_t offset, int bits,
 	case 0163:
 	    length++;
 	    ins->rex |= REX_D;
-	    ins->drexdst = regval(&ins->oprs[c & 3]);
+	    ins->drexdst = regval(opx);
 	    break;
 	case 0164:
 	case 0165:
@@ -938,13 +966,39 @@ static int64_t calcsize(int32_t segment, int64_t offset, int bits,
 	case 0167:
 	    length++;
 	    ins->rex |= REX_D|REX_OC;
-	    ins->drexdst = regval(&ins->oprs[c & 3]);
+	    ins->drexdst = regval(opx);
 	    break;
-        case 0170:
-            length++;
-            break;
 	case 0171:
 	    break;
+	case 0172:
+	case 0173:
+	case 0174:
+	    codes++;
+	    length++;
+	    break;
+        case 0250:
+        case 0251:
+        case 0252:
+        case 0253:
+            length += is_sbyte64(ins, c & 3) ? 1 : 4;
+            break;
+	case 0260:
+	case 0261:
+	case 0262:
+	case 0263:
+	    length += 2;
+	    ins->rex |= REX_V;
+	    ins->drexdst = regval(opx);
+	    ins->vex_m = *codes++;
+	    ins->vex_wlp = *codes++;
+	    break;
+	case 0270:
+	    length += 2;
+	    ins->rex |= REX_V;
+	    ins->drexdst = 0;
+	    ins->vex_m = *codes++;
+	    ins->vex_wlp = *codes++;
+	    break;
         case 0300:
         case 0301:
         case 0302:
@@ -1005,6 +1059,13 @@ static int64_t calcsize(int32_t segment, int64_t offset, int bits,
             else
                 length += ins->oprs[0].offset;
             break;
+	case 0360:
+	    break;
+	case 0361:
+	case 0362:
+	case 0363:
+	    length++;
+	    break;
 	case 0364:
 	case 0365:
 	    break;
@@ -1029,7 +1090,7 @@ static int64_t calcsize(int32_t segment, int64_t offset, int bits,
 		if (c <= 0177) {
 		    /* pick rfield from operand b */
 		    rflags = regflag(&ins->oprs[c & 7]);
-		    rfield = regvals[ins->oprs[c & 7].basereg];
+		    rfield = nasm_regvals[ins->oprs[c & 7].basereg];
 		} else {
 		    rflags = 0;
 		    rfield = c & 7;
@@ -1053,12 +1114,41 @@ static int64_t calcsize(int32_t segment, int64_t offset, int bits,
 
     ins->rex &= rex_mask;
 
-    if (ins->rex & REX_D) {
+    if (ins->rex & REX_V) {
+	int bad32 = REX_R|REX_W|REX_X|REX_B;
+
+	if (ins->rex & REX_H) {
+	    errfunc(ERR_NONFATAL, "cannot use high register in vex instruction");
+	    return -1;
+	}
+	switch (ins->vex_wlp & 030) {
+	case 000:
+	case 020:
+	    ins->rex &= ~REX_W;
+	    break;
+	case 010:
+	    ins->rex |= REX_W;
+	    bad32 &= ~REX_W;
+	    break;
+	case 030:
+	    /* Follow REX_W */
+	    break;
+	}
+
+	if (bits != 64 && ((ins->rex & bad32) || ins->drexdst > 7)) {
+	    errfunc(ERR_NONFATAL, "invalid operands in non-64-bit mode");
+	    return -1;
+	}
+	if (ins->vex_m != 1 || (ins->rex & (REX_W|REX_R|REX_B)))
+	    length += 3;
+	else
+	    length += 2;
+    } else if (ins->rex & REX_D) {
 	if (ins->rex & REX_H) {
 	    errfunc(ERR_NONFATAL, "cannot use high register in drex instruction");
 	    return -1;
 	}
-	if (bits != 64 && ((ins->rex & (REX_W|REX_X|REX_B)) ||
+	if (bits != 64 && ((ins->rex & (REX_R|REX_W|REX_X|REX_B)) ||
 			   ins->drexdst > 7)) {
 	    errfunc(ERR_NONFATAL, "invalid operands in non-64-bit mode");
 	    return -1;
@@ -1086,7 +1176,7 @@ static int64_t calcsize(int32_t segment, int64_t offset, int bits,
 }
 
 #define EMIT_REX()							\
-    if (!(ins->rex & REX_D) && (ins->rex & REX_REAL) && (bits == 64)) {	\
+    if (!(ins->rex & (REX_D|REX_V)) && (ins->rex & REX_REAL) && (bits == 64)) { \
 	ins->rex = (ins->rex & REX_REAL)|REX_P;				\
 	out(offset, segment, &ins->rex, OUT_RAWDATA, 1, NO_SEG, NO_SEG); \
 	ins->rex = 0;							\
@@ -1094,7 +1184,7 @@ static int64_t calcsize(int32_t segment, int64_t offset, int bits,
     }
 
 static void gencode(int32_t segment, int64_t offset, int bits,
-                    insn * ins, const char *codes, int64_t insn_end)
+                    insn * ins, const uint8_t *codes, int64_t insn_end)
 {
     static char condval[] = {   /* conditional opcodes */
         0x7, 0x3, 0x2, 0x6, 0x2, 0x4, 0xF, 0xD, 0xC, 0xE, 0x6, 0x2,
@@ -1174,6 +1264,7 @@ static void gencode(int32_t segment, int64_t offset, int bits,
         case 015:
         case 016:
 	case 017:
+	    /* XXX: warns for legitimate optimizer actions */
             if (opx->offset < -128 || opx->offset > 127) {
                 errfunc(ERR_WARNING | ERR_WARN_NOV,
 			"signed byte value exceeds bounds");
@@ -1371,7 +1462,7 @@ static void gencode(int32_t segment, int64_t offset, int bits,
             if (opx->segment == NO_SEG)
                 errfunc(ERR_NONFATAL, "value referenced by FAR is not"
                         " relocatable");
-            data = 0L;
+	    data = 0;
             out(offset, segment, &data, OUT_ADDRESS, 2,
                 outfmt->segbase(1 + opx->segment),
                 opx->wrt);
@@ -1383,7 +1474,7 @@ static void gencode(int32_t segment, int64_t offset, int bits,
         case 0142:
 	case 0143:
             data = opx->offset;
-            if (is_sbyte(ins, c & 3, 16)) {
+            if (is_sbyte16(ins, c & 3)) {
                 bytes[0] = data;
                 out(offset, segment, bytes, OUT_RAWDATA, 1, NO_SEG,
                     NO_SEG);
@@ -1404,7 +1495,7 @@ static void gencode(int32_t segment, int64_t offset, int bits,
 	case 0147:
 	    EMIT_REX();
             bytes[0] = *codes++;
-            if (is_sbyte(ins, c & 3, 16))
+            if (is_sbyte16(ins, c & 3))
                 bytes[0] |= 2;  /* s-bit */
             out(offset, segment, bytes, OUT_RAWDATA, 1, NO_SEG, NO_SEG);
             offset++;
@@ -1415,7 +1506,7 @@ static void gencode(int32_t segment, int64_t offset, int bits,
         case 0152:
 	case 0153:
             data = opx->offset;
-            if (is_sbyte(ins, c & 3, 32)) {
+            if (is_sbyte32(ins, c & 3)) {
                 bytes[0] = data;
                 out(offset, segment, bytes, OUT_RAWDATA, 1, NO_SEG,
                     NO_SEG);
@@ -1433,7 +1524,7 @@ static void gencode(int32_t segment, int64_t offset, int bits,
 	case 0157:
 	    EMIT_REX();
             bytes[0] = *codes++;
-            if (is_sbyte(ins, c & 3, 32))
+            if (is_sbyte32(ins, c & 3))
                 bytes[0] |= 2;  /* s-bit */
             out(offset, segment, bytes, OUT_RAWDATA, 1, NO_SEG, NO_SEG);
             offset++;
@@ -1449,13 +1540,6 @@ static void gencode(int32_t segment, int64_t offset, int bits,
 	case 0167:
 	    break;
 
-        case 0170:
-	    EMIT_REX();
-            bytes[0] = 0;
-            out(offset, segment, bytes, OUT_RAWDATA, 1, NO_SEG, NO_SEG);
-            offset += 1;
-            break;
-
 	case 0171:
 	    bytes[0] =
 		(ins->drexdst << 4) |
@@ -1466,6 +1550,83 @@ static void gencode(int32_t segment, int64_t offset, int bits,
 	    offset++;
 	    break;
 
+	case 0172:
+	    c = *codes++;
+	    opx = &ins->oprs[c >> 3];
+	    bytes[0] = nasm_regvals[opx->basereg] << 4;
+	    opx = &ins->oprs[c & 7];
+	    if (opx->segment != NO_SEG || opx->wrt != NO_SEG) {
+		errfunc(ERR_NONFATAL,
+			"non-absolute expression not permitted as argument %d",
+			c & 7);
+	    } else {
+		if (opx->offset & ~15) {
+		    errfunc(ERR_WARNING | ERR_WARN_NOV,
+			    "four-bit argument exceeds bounds");
+		}
+		bytes[0] |= opx->offset & 15;
+	    }
+	    out(offset, segment, bytes, OUT_RAWDATA, 1, NO_SEG, NO_SEG);
+	    offset++;
+	    break;
+
+	case 0173:
+	    c = *codes++;
+	    opx = &ins->oprs[c >> 4];
+	    bytes[0] = nasm_regvals[opx->basereg] << 4;
+	    bytes[0] |= c & 15;
+	    out(offset, segment, bytes, OUT_RAWDATA, 1, NO_SEG, NO_SEG);
+	    offset++;
+	    break;
+
+	case 0174:
+	    c = *codes++;
+	    opx = &ins->oprs[c];
+	    bytes[0] = nasm_regvals[opx->basereg] << 4;
+	    out(offset, segment, bytes, OUT_RAWDATA, 1, NO_SEG, NO_SEG);
+	    offset++;
+	    break;
+
+        case 0250:
+        case 0251:
+        case 0252:
+	case 0253:
+            data = opx->offset;
+	    /* is_sbyte32() is right here, we have already warned */
+            if (is_sbyte32(ins, c & 3)) {
+                bytes[0] = data;
+                out(offset, segment, bytes, OUT_RAWDATA, 1, NO_SEG,
+                    NO_SEG);
+                offset++;
+            } else {
+                out(offset, segment, &data, OUT_ADDRESS, 4,
+                    opx->segment, opx->wrt);
+                offset += 4;
+            }
+            break;
+
+	case 0260:
+	case 0261:
+	case 0262:
+	case 0263:
+	case 0270:
+	    codes += 2;
+	    if (ins->vex_m != 1 || (ins->rex & (REX_W|REX_X|REX_B))) {
+		bytes[0] = 0xc4;
+		bytes[1] = ins->vex_m | ((~ins->rex & 7) << 5);
+		bytes[2] = ((ins->rex & REX_W) << (7-3)) |
+		    ((~ins->drexdst & 15)<< 3) | (ins->vex_wlp & 07);
+		out(offset, segment, &bytes, OUT_RAWDATA, 3, NO_SEG, NO_SEG);
+		offset += 3;
+	    } else {
+		bytes[0] = 0xc5;
+		bytes[1] = ((~ins->rex & REX_R) << (7-2)) |
+		    ((~ins->drexdst & 15) << 3) | (ins->vex_wlp & 07);
+		out(offset, segment, &bytes, OUT_RAWDATA, 2, NO_SEG, NO_SEG);
+		offset += 2;
+	    }
+	    break;
+
         case 0300:
         case 0301:
         case 0302:
@@ -1569,6 +1730,22 @@ static void gencode(int32_t segment, int64_t offset, int bits,
             }
             break;
 
+	case 0360:
+	    break;
+
+	case 0361:
+	    bytes[0] = 0x66;
+            out(offset, segment, bytes, OUT_RAWDATA, 1, NO_SEG, NO_SEG);
+            offset += 1;
+	    break;
+
+	case 0362:
+	case 0363:
+	    bytes[0] = c - 0362 + 0xf2;
+            out(offset, segment, bytes, OUT_RAWDATA, 1, NO_SEG, NO_SEG);
+            offset += 1;
+	    break;
+
 	case 0364:
 	case 0365:
 	    break;
@@ -1602,7 +1779,7 @@ static void gencode(int32_t segment, int64_t offset, int bits,
                 if (c <= 0177) {
 		    /* pick rfield from operand b */
 		    rflags = regflag(&ins->oprs[c & 7]);
-                    rfield = regvals[ins->oprs[c & 7].basereg];
+                    rfield = nasm_regvals[ins->oprs[c & 7].basereg];
 		} else {
 		    /* rfield is constant */
 		    rflags = 0;
@@ -1676,7 +1853,7 @@ static int32_t regflag(const operand * o)
     if (o->basereg < EXPR_REG_START || o->basereg >= REG_ENUM_LIMIT) {
         errfunc(ERR_PANIC, "invalid operand passed to regflag()");
     }
-    return reg_flags[o->basereg];
+    return nasm_reg_flags[o->basereg];
 }
 
 static int32_t regval(const operand * o)
@@ -1684,7 +1861,7 @@ static int32_t regval(const operand * o)
     if (o->basereg < EXPR_REG_START || o->basereg >= REG_ENUM_LIMIT) {
         errfunc(ERR_PANIC, "invalid operand passed to regval()");
     }
-    return regvals[o->basereg];
+    return nasm_regvals[o->basereg];
 }
 
 static int op_rexflags(const operand * o, int mask)
@@ -1696,8 +1873,8 @@ static int op_rexflags(const operand * o, int mask)
         errfunc(ERR_PANIC, "invalid operand passed to op_rexflags()");
     }
 
-    flags = reg_flags[o->basereg];
-    val = regvals[o->basereg];
+    flags = nasm_reg_flags[o->basereg];
+    val = nasm_regvals[o->basereg];
 
     return rexflags(val, flags, mask);
 }
@@ -1744,47 +1921,13 @@ static int matches(const struct itemplate *itemp, insn * instruction, int bits)
             return 0;
 
     /*
-     * Check that the operand flags all match up
-     */
-    for (i = 0; i < itemp->operands; i++) {
-	if (itemp->opd[i] & SAME_AS) {
-	    int j = itemp->opd[i] & ~SAME_AS;
-	    if (instruction->oprs[i].type != instruction->oprs[j].type ||
-		instruction->oprs[i].basereg != instruction->oprs[j].basereg)
-		return 0;
-	} else  if (itemp->opd[i] & ~instruction->oprs[i].type ||
-            ((itemp->opd[i] & SIZE_MASK) &&
-             ((itemp->opd[i] ^ instruction->oprs[i].type) & SIZE_MASK))) {
-            if ((itemp->opd[i] & ~instruction->oprs[i].type & ~SIZE_MASK) ||
-                (instruction->oprs[i].type & SIZE_MASK))
-                return 0;
-            else
-                return 1;
-        }
-    }
-
-    /*
-     * Check operand sizes
+     * Process size flags
      */
     if (itemp->flags & IF_ARMASK) {
 	memset(size, 0, sizeof size);
 
-        switch (itemp->flags & IF_ARMASK) {
-        case IF_AR0:
-            i = 0;
-            break;
-        case IF_AR1:
-            i = 1;
-            break;
-        case IF_AR2:
-            i = 2;
-            break;
-	case IF_AR3:
-	    i = 3;
-	    break;
-        default:
-            break;              /* Shouldn't happen */
-        }
+	i = ((itemp->flags & IF_ARMASK) >> IF_ARSHFT) - 1;
+
 	switch (itemp->flags & IF_SMASK) {
 	case IF_SB:
             size[i] = BITS8;
@@ -1801,6 +1944,22 @@ static int matches(const struct itemplate *itemp, insn * instruction, int bits)
 	case IF_SO:
 	    size[i] = BITS128;
 	    break;
+	case IF_SY:
+	    size[i] = BITS256;
+	    break;
+	case IF_SZ:
+	    switch (bits) {
+	    case 16:
+		size[i] = BITS16;
+		break;
+	    case 32:
+		size[i] = BITS32;
+		break;
+	    case 64:
+		size[i] = BITS64;
+		break;
+	    }
+	    break;
 	default:
 	    break;
         }
@@ -1809,23 +1968,34 @@ static int matches(const struct itemplate *itemp, insn * instruction, int bits)
 	switch (itemp->flags & IF_SMASK) {
 	case IF_SB:
             asize = BITS8;
-            oprs = itemp->operands;
 	    break;
 	case IF_SW:
             asize = BITS16;
-            oprs = itemp->operands;
 	    break;
 	case IF_SD:
             asize = BITS32;
-            oprs = itemp->operands;
 	    break;
 	case IF_SQ:
             asize = BITS64;
-            oprs = itemp->operands;
 	    break;
 	case IF_SO:
             asize = BITS128;
-            oprs = itemp->operands;
+	    break;
+	case IF_SY:
+            asize = BITS256;
+	    break;
+	case IF_SZ:
+	    switch (bits) {
+	    case 16:
+		asize = BITS16;
+		break;
+	    case 32:
+		asize = BITS32;
+		break;
+	    case 64:
+		asize = BITS64;
+		break;
+	    }
 	    break;
 	default:
 	    break;
@@ -1834,6 +2004,33 @@ static int matches(const struct itemplate *itemp, insn * instruction, int bits)
 	    size[i] = asize;
     }
 
+    /*
+     * Check that the operand flags all match up
+     */
+    for (i = 0; i < itemp->operands; i++) {
+	int32_t type = instruction->oprs[i].type;
+	if (!(type & SIZE_MASK))
+	    type |= size[i];
+
+	if (itemp->opd[i] & SAME_AS) {
+	    int j = itemp->opd[i] & ~SAME_AS;
+	    if (type != instruction->oprs[j].type ||
+		instruction->oprs[i].basereg != instruction->oprs[j].basereg)
+		return 0;
+	} else if (itemp->opd[i] & ~type ||
+            ((itemp->opd[i] & SIZE_MASK) &&
+             ((itemp->opd[i] ^ type) & SIZE_MASK))) {
+            if ((itemp->opd[i] & ~type & ~SIZE_MASK) ||
+                (type & SIZE_MASK))
+                return 0;
+            else
+                return 1;
+        }
+    }
+
+    /*
+     * Check operand sizes
+     */
     if (itemp->flags & (IF_SM | IF_SM2)) {
         oprs = (itemp->flags & IF_SM2 ? 2 : itemp->operands);
         asize = 0;
@@ -1892,7 +2089,7 @@ static ea *process_ea(operand * input, ea * output, int bits,
             || input->basereg >= REG_ENUM_LIMIT)
             return NULL;
 	f = regflag(input);
-        i = regvals[input->basereg];
+        i = nasm_regvals[input->basereg];
 
 	if (REG_EA & ~f)
 	    return NULL;	/* Invalid EA register */
@@ -1934,16 +2131,16 @@ static ea *process_ea(operand * input, ea * output, int bits,
                 i = -1;         /* make this easy, at least */
 
             if (i >= EXPR_REG_START && i < REG_ENUM_LIMIT) {
-                it = regvals[i];
-		ix = reg_flags[i];
+                it = nasm_regvals[i];
+		ix = nasm_reg_flags[i];
 	    } else {
                 it = -1;
 		ix = 0;
 	    }
 
 	    if (b >= EXPR_REG_START && b < REG_ENUM_LIMIT) {
-                bt = regvals[b];
-		bx = reg_flags[b];
+                bt = nasm_regvals[b];
+		bx = nasm_reg_flags[b];
 	    } else {
                 bt = -1;
 		bx = 0;
@@ -2216,14 +2413,14 @@ static void add_asp(insn *ins, int addrbits)
 		|| ins->oprs[j].indexreg >= REG_ENUM_LIMIT)
 		i = 0;
 	    else
-		i = reg_flags[ins->oprs[j].indexreg];
+		i = nasm_reg_flags[ins->oprs[j].indexreg];
 
 	    /* Verify as Register */
 	    if (ins->oprs[j].basereg < EXPR_REG_START
 		|| ins->oprs[j].basereg >= REG_ENUM_LIMIT)
 		b = 0;
 	    else
-		b = reg_flags[ins->oprs[j].basereg];
+		b = nasm_reg_flags[ins->oprs[j].basereg];
 
 	    if (ins->oprs[j].scale == 0)
 		i = 0;
