@@ -2,7 +2,7 @@
  *
  * The Netwide Assembler is copyright (C) 1996 Simon Tatham and
  * Julian Hall. All rights reserved. The software is
- * redistributable under the licence given in the file "Licence"
+ * redistributable under the license given in the file "LICENSE"
  * distributed in the NASM archive.
  *
  * initial version 27/iii/95 by Simon Tatham
@@ -19,8 +19,8 @@
 #include "disasm.h"
 #include "sync.h"
 #include "insns.h"
-
-#include "names.c"
+#include "tables.h"
+#include "regdis.h"
 
 /*
  * Flags that go into the `segment' field of `insn' structures
@@ -36,8 +36,6 @@
 #define SEG_SIGNED	128
 #define SEG_64BIT	256
 
-#include "regdis.c"
-
 /*
  * Prefix information
  */
@@ -49,11 +47,15 @@ struct prefix_info {
     uint8_t rep;		/* Rep prefix present */
     uint8_t seg;		/* Segment override prefix present */
     uint8_t lock;		/* Lock prefix present */
-    uint8_t rex;		/* Rex prefix present */
+    uint8_t vex[3];		/* VEX prefix present */
+    uint8_t vex_m;		/* VEX.M field */
+    uint8_t vex_v;
+    uint8_t vex_lp;		/* VEX.LP fields */
+    uint32_t rex;		/* REX prefix present */
 };
 
 #define getu8(x) (*(uint8_t *)(x))
-#if defined(__i386__) || defined(__x86_64__)
+#if X86_MEMORY
 /* Littleendian CPU which can handle unaligned references */
 #define getu16(x) (*(uint16_t *)(x))
 #define getu32(x) (*(uint32_t *)(x))
@@ -112,15 +114,19 @@ static enum reg_enum whichreg(int32_t regflags, int regval, int rex)
         return R_RCX;
     if (!(FPU0 & ~regflags))
         return R_ST0;
+    if (!(XMM0 & ~regflags))
+	return R_XMM0;
+    if (!(YMM0 & ~regflags))
+	return R_YMM0;
     if (!(REG_CS & ~regflags))
         return (regval == 1) ? R_CS : 0;
     if (!(REG_DESS & ~regflags))
         return (regval == 0 || regval == 2
-                || regval == 3 ? rd_sreg[regval] : 0);
+                || regval == 3 ? nasm_rd_sreg[regval] : 0);
     if (!(REG_FSGS & ~regflags))
-        return (regval == 4 || regval == 5 ? rd_sreg[regval] : 0);
+        return (regval == 4 || regval == 5 ? nasm_rd_sreg[regval] : 0);
     if (!(REG_SEG67 & ~regflags))
-        return (regval == 6 || regval == 7 ? rd_sreg[regval] : 0);
+        return (regval == 6 || regval == 7 ? nasm_rd_sreg[regval] : 0);
 
     /* All the entries below look up regval in an 16-entry array */
     if (regval < 0 || regval > 15)
@@ -128,46 +134,39 @@ static enum reg_enum whichreg(int32_t regflags, int regval, int rex)
 
     if (!(REG8 & ~regflags)) {
 	if (rex & REX_P)
-	    return rd_reg8_rex[regval];
+	    return nasm_rd_reg8_rex[regval];
 	else
-	    return rd_reg8[regval];
+	    return nasm_rd_reg8[regval];
     }
     if (!(REG16 & ~regflags))
-        return rd_reg16[regval];
+        return nasm_rd_reg16[regval];
     if (!(REG32 & ~regflags))
-        return rd_reg32[regval];
+        return nasm_rd_reg32[regval];
     if (!(REG64 & ~regflags))
-        return rd_reg64[regval];
+        return nasm_rd_reg64[regval];
     if (!(REG_SREG & ~regflags))
-        return rd_sreg[regval & 7]; /* Ignore REX */
+        return nasm_rd_sreg[regval & 7]; /* Ignore REX */
     if (!(REG_CREG & ~regflags))
-        return rd_creg[regval];
+        return nasm_rd_creg[regval];
     if (!(REG_DREG & ~regflags))
-        return rd_dreg[regval];
+        return nasm_rd_dreg[regval];
     if (!(REG_TREG & ~regflags)) {
 	if (rex & REX_P)
 	    return 0;		/* TR registers are ill-defined with rex */
-        return rd_treg[regval];
+        return nasm_rd_treg[regval];
     }
     if (!(FPUREG & ~regflags))
-        return rd_fpureg[regval & 7]; /* Ignore REX */
+        return nasm_rd_fpureg[regval & 7]; /* Ignore REX */
     if (!(MMXREG & ~regflags))
-        return rd_mmxreg[regval & 7]; /* Ignore REX */
+        return nasm_rd_mmxreg[regval & 7]; /* Ignore REX */
     if (!(XMMREG & ~regflags))
-        return rd_xmmreg[regval];
+        return nasm_rd_xmmreg[regval];
+    if (!(YMMREG & ~regflags))
+        return nasm_rd_ymmreg[regval];
 
     return 0;
 }
 
-static const char *whichcond(int condval)
-{
-    static int conds[] = {
-        C_O, C_NO, C_C, C_NC, C_Z, C_NZ, C_NA, C_A,
-        C_S, C_NS, C_PE, C_PO, C_L, C_NL, C_NG, C_G
-    };
-    return conditions[conds[condval]];
-}
-
 /*
  * Process a DREX suffix
  */
@@ -295,9 +294,9 @@ static uint8_t *do_ea(uint8_t *data, int modrm, int asize,
         op->indexreg = -1;
 
 	if (a64)
-	    op->basereg = rd_reg64[rm | ((rex & REX_B) ? 8 : 0)];
+	    op->basereg = nasm_rd_reg64[rm | ((rex & REX_B) ? 8 : 0)];
 	else
-	    op->basereg = rd_reg32[rm | ((rex & REX_B) ? 8 : 0)];
+	    op->basereg = nasm_rd_reg32[rm | ((rex & REX_B) ? 8 : 0)];
 
         if (rm == 5 && mod == 0) {
 	    if (segsize == 64) {
@@ -323,19 +322,19 @@ static uint8_t *do_ea(uint8_t *data, int modrm, int asize,
 	    if (index == 4)
 		op->indexreg = -1; /* ESP/RSP/R12 cannot be an index */
 	    else if (a64)
-		op->indexreg = rd_reg64[index | ((rex & REX_X) ? 8 : 0)];
+		op->indexreg = nasm_rd_reg64[index | ((rex & REX_X) ? 8 : 0)];
 	    else
-		op->indexreg = rd_reg64[index | ((rex & REX_X) ? 8 : 0)];
+		op->indexreg = nasm_rd_reg32[index | ((rex & REX_X) ? 8 : 0)];
 
 	    if (base == 5 && mod == 0) {
 		op->basereg = -1;
 		mod = 2;	/* Fake disp32 */
 	    } else if (a64)
-		op->basereg = rd_reg64[base | ((rex & REX_B) ? 8 : 0)];
+		op->basereg = nasm_rd_reg64[base | ((rex & REX_B) ? 8 : 0)];
 	    else
-		op->basereg = rd_reg32[base | ((rex & REX_B) ? 8 : 0)];
+		op->basereg = nasm_rd_reg32[base | ((rex & REX_B) ? 8 : 0)];
 
-	    if (segsize != 32)
+	    if (segsize == 16)
 		op->disp_size = 32;
         }
 
@@ -350,7 +349,7 @@ static uint8_t *do_ea(uint8_t *data, int modrm, int asize,
             break;
         case 2:
             op->segment |= SEG_DISP32;
-            op->offset = getu32(data);
+            op->offset = gets32(data);
 	    data += 4;
             break;
         }
@@ -362,6 +361,8 @@ static uint8_t *do_ea(uint8_t *data, int modrm, int asize,
  * Determine whether the instruction template in t corresponds to the data
  * stream in data. Return the number of bytes matched if so.
  */
+#define case4(x) case (x): case (x)+1: case (x)+2: case (x)+3
+
 static int matches(const struct itemplate *t, uint8_t *data,
 		   const struct prefix_info *prefix, int segsize, insn *ins)
 {
@@ -372,7 +373,11 @@ static int matches(const struct itemplate *t, uint8_t *data,
     uint8_t lock = prefix->lock;
     int osize = prefix->osize;
     int asize = prefix->asize;
-    int i;
+    int i, c;
+    struct operand *opx;
+    int s_field_for = -1;	/* No 144/154 series code encountered */
+    bool vex_ok = false;
+    int regmask = (segsize == 64) ? 15 : 7;
 
     for (i = 0; i < MAX_OPERANDS; i++) {
 	ins->oprs[i].segment = ins->oprs[i].disp_size =
@@ -380,6 +385,7 @@ static int matches(const struct itemplate *t, uint8_t *data,
     }
     ins->condition = -1;
     ins->rex = prefix->rex;
+    memset(ins->prefixes, 0, sizeof ins->prefixes);
 
     if (t->flags & (segsize == 64 ? IF_NOLONG : IF_LONG))
         return false;
@@ -389,15 +395,19 @@ static int matches(const struct itemplate *t, uint8_t *data,
     else if (prefix->rep == 0xF3)
         drep = P_REP;
 
-    while (*r) {
-        int c = *r++;
+    while ((c = *r++) != 0) {
+	opx = &ins->oprs[c & 3];
 
-	/* FIX: change this into a switch */
-	if (c >= 01 && c <= 03) {
+	switch (c) {
+	case 01:
+	case 02:
+	case 03:
             while (c--)
                 if (*r++ != *data++)
                     return false;
-	} else if (c == 04) {
+	    break;
+
+	case 04:
             switch (*data++) {
             case 0x07:
                 ins->oprs[0].basereg = 0;
@@ -411,7 +421,9 @@ static int matches(const struct itemplate *t, uint8_t *data,
             default:
                 return false;
             }
-	} else if (c == 05) {
+	    break;
+
+	case 05:
             switch (*data++) {
             case 0xA1:
                 ins->oprs[0].basereg = 4;
@@ -422,7 +434,9 @@ static int matches(const struct itemplate *t, uint8_t *data,
             default:
                 return false;
 	    }
-	} else if (c == 06) {
+	    break;
+
+	case 06:
             switch (*data++) {
             case 0x06:
                 ins->oprs[0].basereg = 0;
@@ -439,7 +453,9 @@ static int matches(const struct itemplate *t, uint8_t *data,
             default:
                 return false;
             }
-	} else if (c == 07) {
+	    break;
+
+	case 07:
             switch (*data++) {
             case 0xA0:
                 ins->oprs[0].basereg = 4;
@@ -450,113 +466,221 @@ static int matches(const struct itemplate *t, uint8_t *data,
             default:
                 return false;
             }
-	} else if (c >= 010 && c <= 013) {
+	    break;
+
+	case4(010):
+	{
             int t = *r++, d = *data++;
             if (d < t || d > t + 7)
                 return false;
             else {
-                ins->oprs[c - 010].basereg = (d-t)+
+                opx->basereg = (d-t)+
 		    (ins->rex & REX_B ? 8 : 0);
-                ins->oprs[c - 010].segment |= SEG_RMREG;
+                opx->segment |= SEG_RMREG;
             }
-	} else if (c >= 014 && c <= 017) {
-            ins->oprs[c - 014].offset = (int8_t)*data++;
-            ins->oprs[c - 014].segment |= SEG_SIGNED;
-        } else if (c >= 020 && c <= 023) {
-            ins->oprs[c - 020].offset = *data++;
-	} else if (c >= 024 && c <= 027) {
-            ins->oprs[c - 024].offset = *data++;
-	} else if (c >= 030 && c <= 033) {
-            ins->oprs[c - 030].offset = getu16(data);
+	    break;
+	}
+
+	case4(014):
+            opx->offset = (int8_t)*data++;
+            opx->segment |= SEG_SIGNED;
+	    break;
+
+	case4(020):
+            opx->offset = *data++;
+	    break;
+
+	case4(024):
+            opx->offset = *data++;
+	    break;
+
+	case4(030):
+            opx->offset = getu16(data);
 	    data += 2;
-        } else if (c >= 034 && c <= 037) {
+	    break;
+
+	case4(034):
 	    if (osize == 32) {
-		ins->oprs[c - 034].offset = getu32(data);
+		opx->offset = getu32(data);
 		data += 4;
 	    } else {
-		ins->oprs[c - 034].offset = getu16(data);
+		opx->offset = getu16(data);
 		data += 2;
 	    }
             if (segsize != asize)
-                ins->oprs[c - 034].disp_size = asize;
-        } else if (c >= 040 && c <= 043) {
-            ins->oprs[c - 040].offset = getu32(data);
+                opx->disp_size = asize;
+	    break;
+
+	case4(040):
+            opx->offset = getu32(data);
 	    data += 4;
-        } else if (c >= 044 && c <= 047) {
+	    break;
+
+	case4(044):
 	    switch (asize) {
 	    case 16:
-		ins->oprs[c - 044].offset = getu16(data);
+		opx->offset = getu16(data);
 		data += 2;
+		if (segsize != 16)
+		    opx->disp_size = 16;
 		break;
 	    case 32:
-		ins->oprs[c - 044].offset = getu32(data);
+		opx->offset = getu32(data);
 		data += 4;
+		if (segsize == 16)
+		    opx->disp_size = 32;
 		break;
 	    case 64:
-		ins->oprs[c - 044].offset = getu64(data);
+		opx->offset = getu64(data);
+		opx->disp_size = 64;
 		data += 8;
 		break;
 	    }
-            if (segsize != asize)
-                ins->oprs[c - 044].disp_size = asize;
-        } else if (c >= 050 && c <= 053) {
-            ins->oprs[c - 050].offset = gets8(data++);
-            ins->oprs[c - 050].segment |= SEG_RELATIVE;
-        } else if (c >= 054 && c <= 057) {
-	    ins->oprs[c - 054].offset = getu64(data);
+	    break;
+
+	case4(050):
+            opx->offset = gets8(data++);
+            opx->segment |= SEG_RELATIVE;
+	    break;
+
+	case4(054):
+	    opx->offset = getu64(data);
 	    data += 8;
-	} else if (c >= 060 && c <= 063) {
-            ins->oprs[c - 060].offset = gets16(data);
+	    break;
+
+	case4(060):
+            opx->offset = gets16(data);
 	    data += 2;
-            ins->oprs[c - 060].segment |= SEG_RELATIVE;
-            ins->oprs[c - 060].segment &= ~SEG_32BIT;
-        } else if (c >= 064 && c <= 067) {
-            ins->oprs[c - 064].segment |= SEG_RELATIVE;
+            opx->segment |= SEG_RELATIVE;
+            opx->segment &= ~SEG_32BIT;
+	    break;
+
+	case4(064):
+            opx->segment |= SEG_RELATIVE;
 	    if (osize == 16) {
-		ins->oprs[c - 064].offset = getu16(data);
+		opx->offset = gets16(data);
 		data += 2;
-                ins->oprs[c - 064].segment &= ~(SEG_32BIT|SEG_64BIT);
+                opx->segment &= ~(SEG_32BIT|SEG_64BIT);
 	    } else if (osize == 32) {
-		ins->oprs[c - 064].offset = getu32(data);
+		opx->offset = gets32(data);
 		data += 4;
-                ins->oprs[c - 064].segment &= ~SEG_64BIT;
-                ins->oprs[c - 064].segment |= SEG_32BIT;
+                opx->segment &= ~SEG_64BIT;
+                opx->segment |= SEG_32BIT;
 	    }
             if (segsize != osize) {
-                ins->oprs[c - 064].type =
-                    (ins->oprs[c - 064].type & ~SIZE_MASK)
+                opx->type =
+                    (opx->type & ~SIZE_MASK)
                     | ((osize == 16) ? BITS16 : BITS32);
             }
-        } else if (c >= 070 && c <= 073) {
-            ins->oprs[c - 070].offset = getu32(data);
+	    break;
+
+	case4(070):
+            opx->offset = gets32(data);
 	    data += 4;
-            ins->oprs[c - 070].segment |= SEG_32BIT | SEG_RELATIVE;
-        } else if (c >= 0100 && c < 0140) {
-            int modrm = *data++;
-            ins->oprs[c & 07].segment |= SEG_RMREG;
+            opx->segment |= SEG_32BIT | SEG_RELATIVE;
+	    break;
+
+	case4(0100):
+	case4(0110):
+	case4(0120):
+	case4(0130):
+	{
+	    int modrm = *data++;
+            opx->segment |= SEG_RMREG;
             data = do_ea(data, modrm, asize, segsize,
-			 &ins->oprs[(c >> 3) & 07], ins);
+			 &ins->oprs[(c >> 3) & 3], ins);
 	    if (!data)
 		return false;
-            ins->oprs[c & 07].basereg = ((modrm >> 3)&7)+
+            opx->basereg = ((modrm >> 3)&7)+
 		(ins->rex & REX_R ? 8 : 0);
-        } else if (c >= 0140 && c <= 0143) {
-            ins->oprs[c - 0140].offset = getu16(data);
-	    data += 2;
-        } else if (c >= 0150 && c <= 0153) {
-	    ins->oprs[c - 0150].offset = getu32(data);
-	    data += 4;
-	} else if (c >= 0160 && c <= 0167) {
-	    ins->rex |= (c & 4) ? REX_D|REX_OC : REX_D;
+	    break;
+	}
+
+	case4(0140):
+	    if (s_field_for == (c & 3)) {
+		opx->offset = gets8(data);
+		data++;
+	    } else {
+		opx->offset = getu16(data);
+		data += 2;
+	    }
+	    break;
+
+	case4(0144):
+	case4(0154):
+	    s_field_for = (*data & 0x02) ? c & 3 : -1;
+	    if ((*data++ & ~0x02) != *r++)
+		return false;
+	    break;
+
+	case4(0150):
+	    if (s_field_for == (c & 3)) {
+		opx->offset = gets8(data);
+		data++;
+	    } else {
+		opx->offset = getu32(data);
+		data += 4;
+	    }
+	    break;
+
+	case4(0160):
+	    ins->rex |= REX_D;
 	    ins->drexdst = c & 3;
-        } else if (c == 0170) {
-            if (*data++)
-                return false;
-	} else if (c == 0171) {
+	    break;
+
+	case4(0164):
+	    ins->rex |= REX_D|REX_OC;
+	    ins->drexdst = c & 3;
+	    break;
+
+	case 0171:
 	    data = do_drex(data, ins);
 	    if (!data)
 		return false;
-        } else if (c >= 0200 && c <= 0277) {
+	    break;
+
+	case 0172:
+	{
+	    uint8_t ximm = *data++;
+	    c = *r++;
+	    ins->oprs[c >> 3].basereg = (ximm >> 4) & regmask;
+	    ins->oprs[c >> 3].segment |= SEG_RMREG;
+	    ins->oprs[c & 7].offset = ximm & 15;
+	}
+	break;
+
+	case 0173:
+	{
+	    uint8_t ximm = *data++;
+	    c = *r++;
+
+	    if ((c ^ ximm) & 15)
+		return false;
+
+	    ins->oprs[c >> 4].basereg = (ximm >> 4) & regmask;
+	    ins->oprs[c >> 4].segment |= SEG_RMREG;
+	}
+	break;
+
+	case 0174:
+	{
+	    uint8_t ximm = *data++;
+	    c = *r++;
+
+	    ins->oprs[c].basereg = (ximm >> 4) & regmask;
+	    ins->oprs[c].segment |= SEG_RMREG;
+	}
+	break;
+
+	case4(0200):
+	case4(0204):
+	case4(0210):
+	case4(0214):
+	case4(0220):
+	case4(0224):
+	case4(0230):
+	case4(0234):
+	{
             int modrm = *data++;
             if (((modrm >> 3) & 07) != (c & 07))
                 return false;   /* spare field doesn't match up */
@@ -564,90 +688,258 @@ static int matches(const struct itemplate *t, uint8_t *data,
                          &ins->oprs[(c >> 3) & 07], ins);
 	    if (!data)
 		return false;
-        } else if (c == 0310) {
+	    break;
+	}
+
+	case4(0260):
+	{
+	    int vexm   = *r++;
+	    int vexwlp = *r++;
+	    ins->rex |= REX_V;
+	    if ((prefix->rex & (REX_V|REX_D|REX_P)) != REX_V)
+		return false;
+
+	    if ((vexm & 0x1f) != prefix->vex_m)
+		return false;
+
+	    switch (vexwlp & 030) {
+	    case 000:
+		if (prefix->rex & REX_W)
+		    return false;
+		break;
+	    case 010:
+		if (!(prefix->rex & REX_W))
+		    return false;
+		ins->rex &= ~REX_W;
+		break;
+	    case 020:		/* VEX.W is a don't care */
+		ins->rex &= ~REX_W;
+		break;
+	    case 030:
+		break;
+	    }
+
+	    if ((vexwlp & 007) != prefix->vex_lp)
+		return false;
+
+	    opx->segment |= SEG_RMREG;
+	    opx->basereg = prefix->vex_v;
+	    vex_ok = true;
+	    break;
+	}
+
+	case 0270:
+	{
+	    int vexm   = *r++;
+	    int vexwlp = *r++;
+	    ins->rex |= REX_V;
+	    if ((prefix->rex & (REX_V|REX_D|REX_P)) != REX_V)
+		return false;
+
+	    if ((vexm & 0x1f) != prefix->vex_m)
+		return false;
+
+	    switch (vexwlp & 030) {
+	    case 000:
+		if (ins->rex & REX_W)
+		    return false;
+		break;
+	    case 010:
+		if (!(ins->rex & REX_W))
+		    return false;
+		break;
+	    default:
+		break;		/* Need to do anything special here? */
+	    }
+
+	    if ((vexwlp & 007) != prefix->vex_lp)
+		return false;
+
+	    if (prefix->vex_v != 0)
+		return false;
+
+	    vex_ok = true;
+	    break;
+	}
+
+	case 0310:
             if (asize != 16)
                 return false;
             else
                 a_used = true;
-        } else if (c == 0311) {
+	    break;
+
+	case 0311:
             if (asize == 16)
                 return false;
             else
                 a_used = true;
-        } else if (c == 0312) {
+	    break;
+
+	case 0312:
             if (asize != segsize)
                 return false;
             else
                 a_used = true;
-	} else if (c == 0313) {
+	    break;
+
+	case 0313:
 	    if (asize != 64)
 		return false;
 	    else
 		a_used = true;
-        } else if (c == 0320) {
+	    break;
+
+	case 0314:
+	    if (prefix->rex & REX_B)
+		return false;
+	    break;
+
+	case 0315:
+	    if (prefix->rex & REX_X)
+		return false;
+	    break;
+
+	case 0316:
+	    if (prefix->rex & REX_R)
+		return false;
+	    break;
+
+	case 0317:
+	    if (prefix->rex & REX_W)
+		return false;
+	    break;
+
+	case 0320:
             if (osize != 16)
                 return false;
             else
                 o_used = true;
-        } else if (c == 0321) {
+	    break;
+
+	case 0321:
             if (osize != 32)
                 return false;
             else
                 o_used = true;
-        } else if (c == 0322) {
+	    break;
+
+	case 0322:
             if (osize != (segsize == 16) ? 16 : 32)
                 return false;
             else
                 o_used = true;
-        } else if (c == 0323) {
+	    break;
+
+	case 0323:
 	    ins->rex |= REX_W;	/* 64-bit only instruction */
 	    osize = 64;
-	} else if (c == 0324) {
+	    o_used = true;
+	    break;
+
+	case 0324:
 	    if (!(ins->rex & (REX_P|REX_W)) || osize != 64)
 		return false;
-	} else if (c == 0330) {
+	    o_used = true;
+	    break;
+
+	case 0330:
+	{
             int t = *r++, d = *data++;
             if (d < t || d > t + 15)
                 return false;
             else
                 ins->condition = d - t;
-        } else if (c == 0331) {
+	    break;
+	}
+
+	case 0331:
             if (prefix->rep)
                 return false;
-	} else if (c == 0332) {
+	    break;
+
+	case 0332:
 	    if (prefix->rep != 0xF2)
 		return false;
-        } else if (c == 0333) {
+	    drep = 0;
+	    break;
+
+	case 0333:
             if (prefix->rep != 0xF3)
                 return false;
             drep = 0;
-        } else if (c == 0334) {
+	    break;
+
+	case 0334:
 	    if (lock) {
 		ins->rex |= REX_R;
 		lock = 0;
 	    }
-        } else if (c == 0335) {
+	    break;
+
+	case 0335:
             if (drep == P_REP)
                 drep = P_REPE;
-	} else if (c == 0364) {
+	    break;
+
+	case 0340:
+	    return false;
+
+	case 0360:
+	    if (prefix->osp || prefix->rep)
+		return false;
+	    break;
+
+	case 0361:
+	    if (!prefix->osp || prefix->rep)
+		return false;
+	    o_used = true;
+	    break;
+
+	case 0362:
+	    if (prefix->osp || prefix->rep != 0xf2)
+		return false;
+	    drep = 0;
+	    break;
+
+	case 0363:
+	    if (prefix->osp || prefix->rep != 0xf3)
+		return false;
+	    drep = 0;
+	    break;
+
+	case 0364:
 	    if (prefix->osp)
 		return false;
-	} else if (c == 0365) {
+	    break;
+
+	case 0365:
 	    if (prefix->asp)
 		return false;
-	} else if (c == 0366) {
+	    break;
+
+	case 0366:
 	    if (!prefix->osp)
 		return false;
 	    o_used = true;
-	} else if (c == 0367) {
+	    break;
+
+	case 0367:
 	    if (!prefix->asp)
 		return false;
-	    o_used = true;
+	    a_used = true;
+	    break;
+
+	default:
+	    return false;	/* Unknown code */
 	}
     }
 
-    /* REX cannot be combined with DREX */
-    if ((ins->rex & REX_D) && (prefix->rex))
+    if (!vex_ok && (ins->rex & REX_V))
+	return false;
+
+    /* REX cannot be combined with DREX or VEX */
+    if ((ins->rex & (REX_D|REX_V)) && (prefix->rex & REX_P))
 	return false;
 
     /*
@@ -668,10 +960,26 @@ static int matches(const struct itemplate *t, uint8_t *data,
 	    return false;
         ins->prefixes[PPS_LREP] = drep;
     }
-    if (!o_used && osize == ((segsize == 16) ? 32 : 16)) {
-	if (ins->prefixes[PPS_OSIZE])
-	    return false;
-        ins->prefixes[PPS_OSIZE] = osize == 16 ? P_O16 : P_O32;
+    if (!o_used) {
+	if (osize != ((segsize == 16) ? 16 : 32)) {
+	    enum prefixes pfx = 0;
+
+	    switch (osize) {
+	    case 16:
+		pfx = P_O16;
+		break;
+	    case 32:
+		pfx = P_O32;
+		break;
+	    case 64:
+		pfx = P_O64;
+		break;
+	    }
+
+	    if (ins->prefixes[PPS_OSIZE])
+		return false;
+	    ins->prefixes[PPS_OSIZE] = pfx;
+	}
     }
     if (!a_used && asize != segsize) {
 	if (ins->prefixes[PPS_ASIZE])
@@ -684,6 +992,12 @@ static int matches(const struct itemplate *t, uint8_t *data,
     return data - origdata;
 }
 
+/* Condition names for disassembly, sorted by x86 code */
+static const char * const condition_name[16] = {
+    "o", "no", "c", "nc", "z", "nz", "na", "a",
+    "s", "ns", "pe", "po", "l", "nl", "ng", "g"
+};
+
 int32_t disasm(uint8_t *data, char *output, int outbufsize, int segsize,
             int32_t offset, int autosync, uint32_t prefer)
 {
@@ -699,6 +1013,7 @@ int32_t disasm(uint8_t *data, char *output, int outbufsize, int segsize,
     uint32_t goodness, best;
     int best_pref;
     struct prefix_info prefix;
+    bool end_prefix;
 
     memset(&ins, 0, sizeof ins);
 
@@ -710,36 +1025,102 @@ int32_t disasm(uint8_t *data, char *output, int outbufsize, int segsize,
     prefix.osize = (segsize == 64) ? 32 : segsize;
     segover = NULL;
     origdata = data;
-    for (;;) {
-        if (*data == 0xF3 || *data == 0xF2)
+
+    ix = itable;
+
+    end_prefix = false;
+    while (!end_prefix) {
+	switch (*data) {
+	case 0xF2:
+	case 0xF3:
             prefix.rep = *data++;
-        else if (*data == 0xF0)
+	    break;
+
+	case 0xF0:
             prefix.lock = *data++;
-        else if (*data == 0x2E)
+	    break;
+
+	case 0x2E:
 	    segover = "cs", prefix.seg = *data++;
-	else if (*data == 0x36)
+	    break;
+	case 0x36:
 	    segover = "ss", prefix.seg = *data++;
-	else if (*data == 0x3E)
+	    break;
+	case 0x3E:
 	    segover = "ds", prefix.seg = *data++;
-	else if (*data == 0x26)
+	    break;
+	case 0x26:
 	    segover = "es", prefix.seg = *data++;
-	else if (*data == 0x64)
+	    break;
+	case 0x64:
 	    segover = "fs", prefix.seg = *data++;
-	else if (*data == 0x65)
+	    break;
+	case 0x65:
 	    segover = "gs", prefix.seg = *data++;
-	else if (*data == 0x66) {
+	    break;
+
+	case 0x66:
 	    prefix.osize = (segsize == 16) ? 32 : 16;
 	    prefix.osp = *data++;
-	} else if (*data == 0x67) {
+	    break;
+	case 0x67:
 	    prefix.asize = (segsize == 32) ? 16 : 32;
 	    prefix.asp = *data++;
-	} else if (segsize == 64 && (*data & 0xf0) == REX_P) {
-	    prefix.rex = *data++;
-	    if (prefix.rex & REX_W)
-		prefix.osize = 64;
-	    break;		/* REX is always the last prefix */
-	} else {
-            break;
+	    break;
+
+	case 0xC4:
+	case 0xC5:
+	    if (segsize == 64 || (data[1] & 0xc0) == 0xc0) {
+		prefix.vex[0] = *data++;
+		prefix.vex[1] = *data++;
+		if (prefix.vex[0] == 0xc4)
+		    prefix.vex[2] = *data++;
+	    }
+	    prefix.rex = REX_V;
+	    if (prefix.vex[0] == 0xc4) {
+		prefix.rex |= (~prefix.vex[1] >> 5) & 7; /* REX_RXB */
+		prefix.rex |= (prefix.vex[2] >> (7-3)) & REX_W;
+		prefix.vex_m = prefix.vex[1] & 0x1f;
+		prefix.vex_v = (~prefix.vex[2] >> 3) & 15;
+		prefix.vex_lp = prefix.vex[2] & 7;
+	    } else {
+		prefix.rex |= (~prefix.vex[1] >> (7-2)) & REX_R;
+		prefix.vex_m = 1;
+		prefix.vex_v = (~prefix.vex[1] >> 3) & 15;
+		prefix.vex_lp = prefix.vex[1] & 7;
+	    }
+
+	    ix = itable_VEX[prefix.vex_m][prefix.vex_lp];
+	    end_prefix = true;
+	    break;
+
+	case REX_P + 0x0:
+	case REX_P + 0x1:
+	case REX_P + 0x2:
+	case REX_P + 0x3:
+	case REX_P + 0x4:
+	case REX_P + 0x5:
+	case REX_P + 0x6:
+	case REX_P + 0x7:
+	case REX_P + 0x8:
+	case REX_P + 0x9:
+	case REX_P + 0xA:
+	case REX_P + 0xB:
+	case REX_P + 0xC:
+	case REX_P + 0xD:
+	case REX_P + 0xE:
+	case REX_P + 0xF:
+	    if (segsize == 64) {
+		prefix.rex = *data++;
+		if (prefix.rex & REX_W)
+		    prefix.osize = 64;
+	    }
+	    end_prefix = true;
+	    break;
+
+	default:
+	    end_prefix = true;
+	    break;
 	}
     }
 
@@ -747,8 +1128,11 @@ int32_t disasm(uint8_t *data, char *output, int outbufsize, int segsize,
     best_p = NULL;
     best_pref = INT_MAX;
 
+    if (!ix)
+	return 0;		/* No instruction table at all... */
+
     dp = data;
-    ix = itable + *dp++;
+    ix += *dp++;
     while (ix->n == -1) {
 	ix = (const struct disasm_index *)ix->p + *dp++;
     }
@@ -764,14 +1148,17 @@ int32_t disasm(uint8_t *data, char *output, int outbufsize, int segsize,
             for (i = 0; i < (*p)->operands; i++) {
                 if (!((*p)->opd[i] & SAME_AS) &&
 		    (
-			/* If it's a mem-only EA but we have a register, die. */
+			/* If it's a mem-only EA but we have a
+			   register, die. */
 			((tmp_ins.oprs[i].segment & SEG_RMREG) &&
 			 !(MEMORY & ~(*p)->opd[i])) ||
-			/* If it's a reg-only EA but we have a memory ref, die. */
+			/* If it's a reg-only EA but we have a memory
+			   ref, die. */
 			(!(tmp_ins.oprs[i].segment & SEG_RMREG) &&
 			 !(REG_EA & ~(*p)->opd[i]) &&
 			 !((*p)->opd[i] & REG_SMASK)) ||
-			/* Register type mismatch (eg FS vs REG_DESS): die. */
+			/* Register type mismatch (eg FS vs REG_DESS):
+			   die. */
 			((((*p)->opd[i] & (REGISTER | FPUREG)) ||
 			  (tmp_ins.oprs[i].segment & SEG_RMREG)) &&
 			 !whichreg((*p)->opd[i],
@@ -846,27 +1233,30 @@ int32_t disasm(uint8_t *data, char *output, int outbufsize, int segsize,
         case P_A32:
             slen += snprintf(output + slen, outbufsize - slen, "a32 ");
             break;
+        case P_A64:
+            slen += snprintf(output + slen, outbufsize - slen, "a64 ");
+            break;
         case P_O16:
             slen += snprintf(output + slen, outbufsize - slen, "o16 ");
             break;
         case P_O32:
             slen += snprintf(output + slen, outbufsize - slen, "o32 ");
             break;
+        case P_O64:
+            slen += snprintf(output + slen, outbufsize - slen, "o64 ");
+            break;
 	default:
 	    break;
         }
 
-    for (i = 0; i < (int)elements(ico); i++)
-        if ((*p)->opcode == ico[i]) {
-            slen +=
-                snprintf(output + slen, outbufsize - slen, "%s%s", icn[i],
-                         whichcond(ins.condition));
-            break;
-        }
-    if (i >= (int)elements(ico))
-        slen +=
-            snprintf(output + slen, outbufsize - slen, "%s",
-                     insn_names[(*p)->opcode]);
+    i = (*p)->opcode;
+    if (i >= FIRST_COND_OPCODE)
+	slen += snprintf(output + slen, outbufsize - slen, "%s%s",
+			 nasm_insn_names[i], condition_name[ins.condition]);
+    else
+        slen += snprintf(output + slen, outbufsize - slen, "%s",
+			 nasm_insn_names[i]);
+
     colon = false;
     length += data - origdata;  /* fix up for prefixes */
     for (i = 0; i < (*p)->operands; i++) {
@@ -889,6 +1279,9 @@ int32_t disasm(uint8_t *data, char *output, int outbufsize, int segsize,
              */
             if (!(o->segment & (SEG_32BIT|SEG_64BIT)))
 		offs &= 0xffff;
+	    else if (segsize != 64)
+		offs &= 0xffffffff;
+
             /*
              * add sync marker, if autosync is on
              */
@@ -908,7 +1301,7 @@ int32_t disasm(uint8_t *data, char *output, int outbufsize, int segsize,
             if (t & TO)
                 slen += snprintf(output + slen, outbufsize - slen, "to ");
             slen += snprintf(output + slen, outbufsize - slen, "%s",
-                             reg_names[reg - EXPR_REG_START]);
+                             nasm_reg_names[reg-EXPR_REG_START]);
         } else if (!(UNITY & ~t)) {
             output[slen++] = '1';
         } else if (t & IMMEDIATE) {
@@ -943,12 +1336,13 @@ int32_t disasm(uint8_t *data, char *output, int outbufsize, int segsize,
                          offs);
         } else if (!(MEM_OFFS & ~t)) {
             slen +=
-                snprintf(output + slen, outbufsize - slen, "[%s%s%s0x%"PRIx64"]",
+                snprintf(output + slen, outbufsize - slen,
+			 "[%s%s%s0x%"PRIx64"]",
                          (segover ? segover : ""),
                          (segover ? ":" : ""),
-                         (o->disp_size ==
-                          32 ? "dword " : o->disp_size ==
-                          16 ? "word " : ""), offs);
+			 (o->disp_size == 64 ? "qword " :
+			  o->disp_size == 32 ? "dword " :
+			  o->disp_size == 16 ? "word " : ""), offs);
             segover = NULL;
         } else if (!(REGMEM & ~t)) {
             int started = false;
@@ -967,6 +1361,12 @@ int32_t disasm(uint8_t *data, char *output, int outbufsize, int segsize,
             if (t & BITS80)
                 slen +=
                     snprintf(output + slen, outbufsize - slen, "tword ");
+            if (t & BITS128)
+                slen +=
+                    snprintf(output + slen, outbufsize - slen, "oword ");
+            if (t & BITS256)
+                slen +=
+                    snprintf(output + slen, outbufsize - slen, "yword ");
             if (t & FAR)
                 slen += snprintf(output + slen, outbufsize - slen, "far ");
             if (t & NEAR)
@@ -989,46 +1389,63 @@ int32_t disasm(uint8_t *data, char *output, int outbufsize, int segsize,
             }
             if (o->basereg != -1) {
                 slen += snprintf(output + slen, outbufsize - slen, "%s",
-                                 reg_names[(o->basereg -
-                                            EXPR_REG_START)]);
+                                 nasm_reg_names[(o->basereg-EXPR_REG_START)]);
                 started = true;
             }
             if (o->indexreg != -1) {
                 if (started)
                     output[slen++] = '+';
                 slen += snprintf(output + slen, outbufsize - slen, "%s",
-                                 reg_names[(o->indexreg -
-                                            EXPR_REG_START)]);
+                                 nasm_reg_names[(o->indexreg-EXPR_REG_START)]);
                 if (o->scale > 1)
                     slen +=
                         snprintf(output + slen, outbufsize - slen, "*%d",
                                  o->scale);
                 started = true;
             }
+
+
             if (o->segment & SEG_DISP8) {
-		int minus = 0;
-		int8_t offset = offs;
-		if (offset < 0) {
-		    minus = 1;
+		const char *prefix;
+		uint8_t offset = offs;
+		if ((int8_t)offset < 0) {
+		    prefix = "-";
 		    offset = -offset;
+		} else {
+		    prefix = "+";
 		}
                 slen +=
                     snprintf(output + slen, outbufsize - slen, "%s0x%"PRIx8"",
-			     minus ? "-" : "+", offset);
+			     prefix, offset);
             } else if (o->segment & SEG_DISP16) {
-		int minus = 0;
-		int16_t offset = offs;
-		if (offset < 0) {
-		    minus = 1;
+		const char *prefix;
+		uint16_t offset = offs;
+		if ((int16_t)offset < 0 && started) {
 		    offset = -offset;
+		    prefix = "-";
+		} else {
+		    prefix = started ? "+" : "";
 		}
                 slen +=
-                    snprintf(output + slen, outbufsize - slen, "%s0x%"PRIx16"",
-			     minus ? "-" : started ? "+" : "", offset);
+                    snprintf(output + slen, outbufsize - slen,
+			     "%s0x%"PRIx16"", prefix, offset);
             } else if (o->segment & SEG_DISP32) {
-		    char *prefix = "";
-		    int32_t offset = offs;
-		    if (offset < 0) {
+		if (prefix.asize == 64) {
+		    const char *prefix;
+		    uint64_t offset = (int64_t)(int32_t)offs;
+		    if ((int32_t)offs < 0 && started) {
+			offset = -offset;
+			prefix = "-";
+		    } else {
+			prefix = started ? "+" : "";
+		    }
+		    slen +=
+			snprintf(output + slen, outbufsize - slen,
+				 "%s0x%"PRIx64"", prefix, offset);
+		} else {
+		    const char *prefix;
+		    uint32_t offset = offs;
+		    if ((int32_t) offset < 0 && started) {
 			offset = -offset;
 			prefix = "-";
 		    } else {
@@ -1037,6 +1454,7 @@ int32_t disasm(uint8_t *data, char *output, int outbufsize, int segsize,
 		    slen +=
 			snprintf(output + slen, outbufsize - slen,
 				 "%s0x%"PRIx32"", prefix, offset);
+		}
             }
             output[slen++] = ']';
         } else {
