@@ -2,7 +2,7 @@
  *
  * The Netwide Assembler is copyright (C) 1996 Simon Tatham and
  * Julian Hall. All rights reserved. The software is
- * redistributable under the licence given in the file "Licence"
+ * redistributable under the license given in the file "LICENSE"
  * distributed in the NASM archive.
  *
  * initial version 27/iii/95 by Simon Tatham
@@ -666,7 +666,9 @@ static expr *expr6(int critical)
 {
     int32_t type;
     expr *e;
-    int32_t label_seg, label_ofs;
+    int32_t label_seg;
+    int64_t label_ofs;
+    char *scope;
 
     switch (i) {
     case '-':
@@ -741,6 +743,7 @@ static expr *expr6(int critical)
     case TOKEN_NUM:
     case TOKEN_REG:
     case TOKEN_ID:
+    case TOKEN_INSN:		/* Opcodes that occur here are really labels */
     case TOKEN_HERE:
     case TOKEN_BASE:
         begintemp();
@@ -754,6 +757,7 @@ static expr *expr6(int critical)
                 hint->base = tokval->t_integer, hint->type = EAH_MAKEBASE;
             break;
         case TOKEN_ID:
+	case TOKEN_INSN:
         case TOKEN_HERE:
         case TOKEN_BASE:
             /*
@@ -764,8 +768,9 @@ static expr *expr6(int critical)
             if (!location->known) {
                 error(ERR_NONFATAL,
                       "%s not supported in preprocess-only mode",
-                      (i == TOKEN_ID ? "symbol references" :
-                       i == TOKEN_HERE ? "`$'" : "`$$'"));
+		      (i == TOKEN_HERE ? "`$'" :
+		       i == TOKEN_BASE ? "`$$'" :
+		       "symbol references"));
                 addtotemp(EXPR_UNKNOWN, 1L);
                 break;
             }
@@ -779,14 +784,15 @@ static expr *expr6(int critical)
                 label_ofs = in_abs_seg ? abs_offset : location->offset;
             } else {
                 if (!labelfunc(tokval->t_charptr, &label_seg, &label_ofs)) {
+                    scope = local_scope(tokval->t_charptr);
                     if (critical == 2) {
-                        error(ERR_NONFATAL, "symbol `%s' undefined",
-                              tokval->t_charptr);
+                        error(ERR_NONFATAL, "symbol `%s%s' undefined",
+                              scope,tokval->t_charptr);
                         return NULL;
                     } else if (critical == 1) {
                         error(ERR_NONFATAL,
-                              "symbol `%s' not defined before use",
-                              tokval->t_charptr);
+                              "symbol `%s%s' not defined before use",
+                              scope,tokval->t_charptr);
                         return NULL;
                     } else {
                         if (opflags)
