@@ -2,7 +2,7 @@
  *
  * The Netwide Assembler is copyright (C) 1996 Simon Tatham and
  * Julian Hall. All rights reserved. The software is
- * redistributable under the licence given in the file "Licence"
+ * redistributable under the license given in the file "LICENSE"
  * distributed in the NASM archive.
  *
  * initial version 13/ix/96 by Simon Tatham
@@ -45,15 +45,6 @@ typedef uint64_t fp_2limb;
 #define LIMB_ALL_BYTES	((fp_limb)0x01010101)
 #define LIMB_BYTE(x)	((x)*LIMB_ALL_BYTES)
 
-#if defined(__i386__) || defined(__x86_64__)
-#define put(a,b) (*(uint32_t *)(a) = (b))
-#else
-#define put(a,b) (((a)[0] = (b)),		\
-		  ((a)[1] = (b) >> 8),		\
-		  ((a)[2] = (b) >> 16),		\
-		  ((a)[3] = (b) >> 24))
-#endif
-
 /* 112 bits + 64 bits for accuracy + 16 bits for rounding */
 #define MANT_LIMBS 6
 
@@ -165,7 +156,7 @@ static int32_t read_exponent(const char *string, int32_t max)
 	} else if (*string == '_') {
 	    /* do nothing */
 	} else {
-	    error(ERR_NONFATAL,
+	    error(ERR_NONFATAL|ERR_PASS1,
 		  "invalid character in floating-point constant %s: '%c'",
 		  "exponent", *string);
 	    return INT32_MAX;
@@ -191,16 +182,18 @@ static bool ieee_flconvert(const char *string, fp_limb *mant,
     int32_t tenpwr, twopwr;
     int32_t extratwos;
     bool started, seendot, warned;
+
+    warned = false;
     p = digits;
     tenpwr = 0;
     started = seendot = false;
-    warned = (pass0 != 1);
+
     while (*string && *string != 'E' && *string != 'e') {
         if (*string == '.') {
             if (!seendot) {
                 seendot = true;
             } else {
-                error(ERR_NONFATAL,
+                error(ERR_NONFATAL|ERR_PASS1,
                       "too many periods in floating-point constant");
                 return false;
             }
@@ -215,7 +208,7 @@ static bool ieee_flconvert(const char *string, fp_limb *mant,
                     *p++ = *string - '0';
                 } else {
                     if (!warned) {
-                        error(ERR_WARNING|ERR_WARN_FL_TOOLONG,
+                        error(ERR_WARNING|ERR_WARN_FL_TOOLONG|ERR_PASS1,
                               "floating-point constant significand contains "
                               "more than %i digits", MANT_DIGITS);
                         warned = true;
@@ -228,7 +221,7 @@ static bool ieee_flconvert(const char *string, fp_limb *mant,
         } else if (*string == '_') {
             /* do nothing */
         } else {
-            error(ERR_NONFATAL,
+            error(ERR_NONFATAL|ERR_PASS1,
                   "invalid character in floating-point constant %s: '%c'",
                   "significand", *string);
             return false;
@@ -514,7 +507,7 @@ static bool ieee_flconvert_bin(const char *string, int bits,
             if (!seendot)
                 seendot = true;
             else {
-                error(ERR_NONFATAL,
+                error(ERR_NONFATAL|ERR_PASS1,
                       "too many periods in floating-point constant");
                 return false;
             }
@@ -556,7 +549,7 @@ static bool ieee_flconvert_bin(const char *string, int bits,
 	} else if (c == '_') {
 	    /* ignore */
         } else {
-            error(ERR_NONFATAL,
+            error(ERR_NONFATAL|ERR_PASS1,
                   "floating-point constant: `%c' is invalid character", c);
             return false;
         }
@@ -648,7 +641,7 @@ enum floats {
 static int to_float(const char *str, int s, uint8_t * result,
                     const struct ieee_format *fmt)
 {
-    fp_limb mant[MANT_LIMBS], *mp, m;
+    fp_limb mant[MANT_LIMBS];
     int32_t exponent = 0;
     int32_t expmax = 1 << (fmt->exponent - 1);
     fp_limb one_mask = LIMB_TOP_BIT >>
@@ -680,7 +673,7 @@ static int to_float(const char *str, int s, uint8_t * result,
 	    type = FL_INFINITY;
             break;
 	default:
-	    error(ERR_NONFATAL,
+	    error(ERR_NONFATAL|ERR_PASS1,
 		  "internal error: unknown FP constant token `%s'\n", str);
 	    type = FL_QNAN;
 	    break;
@@ -726,7 +719,7 @@ static int to_float(const char *str, int s, uint8_t * result,
 		type = FL_NORMAL;
             } else if (exponent > 0) {
 		if (pass0 == 1)
-		    error(ERR_WARNING|ERR_WARN_FL_OVERFLOW,
+		    error(ERR_WARNING|ERR_WARN_FL_OVERFLOW|ERR_PASS1,
 			  "overflow in floating-point constant");
 		type = FL_INFINITY;
 	    } else {
@@ -761,14 +754,12 @@ static int to_float(const char *str, int s, uint8_t * result,
 	} else {
 	    if (daz || is_zero(mant)) {
 		/* Flush denormals to zero */
-		if (pass0 == 1)
-		    error(ERR_WARNING|ERR_WARN_FL_UNDERFLOW,
-			  "underflow in floating-point constant");
+		error(ERR_WARNING|ERR_WARN_FL_UNDERFLOW|ERR_PASS1,
+		      "underflow in floating-point constant");
 		goto zero;
 	    } else {
-		if (pass0 == 1)
-		    error(ERR_WARNING|ERR_WARN_FL_DENORM,
-			  "denormal floating-point constant");
+		error(ERR_WARNING|ERR_WARN_FL_DENORM|ERR_PASS1,
+		      "denormal floating-point constant");
 	    }
 	}
 	break;
@@ -783,8 +774,7 @@ static int to_float(const char *str, int s, uint8_t * result,
 	    ieee_shr(mant, 1);
 	    exponent++;
 	    if (exponent >= (expmax << 1)-1) {
-		if (pass0 == 1)
-		    error(ERR_WARNING|ERR_WARN_FL_OVERFLOW,
+		    error(ERR_WARNING|ERR_WARN_FL_OVERFLOW|ERR_PASS1,
 			  "overflow in floating-point constant");
 		type = FL_INFINITY;
 		goto overflow;
@@ -814,21 +804,13 @@ static int to_float(const char *str, int s, uint8_t * result,
 
     mant[0] |= minus ? LIMB_TOP_BIT : 0;
 
-    m = mant[fmt->bytes/LIMB_BYTES];
-    for (i = LIMB_BYTES-(fmt->bytes % LIMB_BYTES); i < LIMB_BYTES; i++)
-	*result++ = m >> (i*8);
-
-    for (mp = &mant[fmt->bytes/LIMB_BYTES], i = 0;
-	 i < fmt->bytes; i += LIMB_BYTES) {
-        m = *--mp;
-        put(result, m);
-        result += LIMB_BYTES;
-    }
+    for (i = fmt->bytes - 1; i >= 0; i--)
+	*result++ = mant[i/LIMB_BYTES] >> (((LIMB_BYTES-1)-(i%LIMB_BYTES))*8);
 
     return 1;                   /* success */
 }
 
-int float_const(const char *number, int32_t sign, uint8_t * result,
+int float_const(const char *number, int sign, uint8_t * result,
                 int bytes, efunc err)
 {
     error = err;
