@@ -2,7 +2,7 @@
  *
  * The Netwide Assembler is copyright (C) 1996 Simon Tatham and
  * Julian Hall. All rights reserved. The software is
- * redistributable under the licence given in the file "Licence"
+ * redistributable under the license given in the file "LICENSE"
  * distributed in the NASM archive.
  *
  * initial version: 27/iii/95 by Simon Tatham
@@ -51,6 +51,32 @@
  */
 struct ofmt;
 
+/*
+ * values for the `type' parameter to an output function.
+ *
+ * Exceptions are OUT_RELxADR, which denote an x-byte relocation
+ * which will be a relative jump. For this we need to know the
+ * distance in bytes from the start of the relocated record until
+ * the end of the containing instruction. _This_ is what is stored
+ * in the size part of the parameter, in this case.
+ *
+ * Also OUT_RESERVE denotes reservation of N bytes of BSS space,
+ * and the contents of the "data" parameter is irrelevant.
+ *
+ * The "data" parameter for the output function points to a "int32_t",
+ * containing the address in question, unless the type is
+ * OUT_RAWDATA, in which case it points to an "uint8_t"
+ * array.
+ */
+enum out_type {
+    OUT_RAWDATA,		/* Plain bytes */
+    OUT_ADDRESS,		/* An address (symbol value) */
+    OUT_RESERVE,		/* Reserved bytes (RESB et al) */
+    OUT_REL2ADR,		/* 2-byte relative address */
+    OUT_REL4ADR,		/* 4-byte relative address */
+    OUT_REL8ADR,		/* 8-byte relative address */
+};
+
 /*
  * -----------------------
  * Other function typedefs
@@ -60,7 +86,7 @@ struct ofmt;
 /*
  * A label-lookup function should look like this.
  */
-typedef bool (*lfunc) (char *label, int32_t *segment, int32_t *offset);
+typedef bool (*lfunc) (char *label, int32_t *segment, int64_t *offset);
 
 /*
  * And a label-definition function like this. The boolean parameter
@@ -68,7 +94,7 @@ typedef bool (*lfunc) (char *label, int32_t *segment, int32_t *offset);
  * should affect the local-label system), or something odder like
  * an EQU or a segment-base symbol, which shouldn't.
  */
-typedef void (*ldfunc) (char *label, int32_t segment, int32_t offset,
+typedef void (*ldfunc) (char *label, int32_t segment, int64_t offset,
                         char *special, bool is_norm, bool isextrn,
                         struct ofmt * ofmt, efunc error);
 
@@ -95,12 +121,12 @@ typedef struct {
      * output-format interface, only OUT_ADDRESS will _always_ be
      * displayed as if it's relocatable, so ensure that any non-
      * relocatable address has been converted to OUT_RAWDATA by
-     * then. Note that OUT_RAWDATA+0 is a valid data type, and is a
+     * then. Note that OUT_RAWDATA,0 is a valid data type, and is a
      * dummy call used to give the listing generator an offset to
      * work with when doing things like uplevel(LIST_TIMES) or
      * uplevel(LIST_INCBIN).
      */
-    void (*output) (int32_t, const void *, uint32_t);
+    void (*output) (int32_t, const void *, enum out_type, uint64_t);
 
     /*
      * Called to send a text line to the listing generator. The
@@ -131,27 +157,11 @@ typedef struct {
     void (*downlevel) (int);
 } ListGen;
 
-/*
- * The expression evaluator must be passed a scanner function; a
- * standard scanner is provided as part of nasmlib.c. The
- * preprocessor will use a different one. Scanners, and the
- * token-value structures they return, look like this.
- *
- * The return value from the scanner is always a copy of the
- * `t_type' field in the structure.
- */
-struct tokenval {
-    int t_type;
-    int64_t t_integer, t_inttwo;
-    char *t_charptr;
-};
-typedef int (*scanner) (void *private_data, struct tokenval * tv);
-
 /*
  * Token types returned by the scanner, in addition to ordinary
  * ASCII character values, and zero for end-of-string.
  */
-enum {                          /* token types, other than chars */
+enum token_type {		/* token types, other than chars */
     TOKEN_INVALID = -1,         /* a placeholder value */
     TOKEN_EOS = 0,              /* end of string */
     TOKEN_EQ = '=', TOKEN_GT = '>', TOKEN_LT = '<',     /* aliases */
@@ -180,6 +190,22 @@ enum floatize {
     FLOAT_128H,
 };
 
+/*
+ * The expression evaluator must be passed a scanner function; a
+ * standard scanner is provided as part of nasmlib.c. The
+ * preprocessor will use a different one. Scanners, and the
+ * token-value structures they return, look like this.
+ *
+ * The return value from the scanner is always a copy of the
+ * `t_type' field in the structure.
+ */
+struct tokenval {
+    enum token_type t_type;
+    char *t_charptr;
+    int64_t t_integer, t_inttwo;
+};
+typedef int (*scanner) (void *private_data, struct tokenval * tv);
+
 struct location {
     int64_t offset;
     int32_t segment;
@@ -265,7 +291,15 @@ typedef expr *(*evalfunc) (scanner sc, void *scprivate,
 #define EXPR_SEGBASE	(EXPR_REG_END+4)
 
 /*
- * Preprocessors ought to look like this:
+ * Linked list of strings...
+ */
+typedef struct string_list {
+    struct string_list *next;
+    char str[1];
+} StrList;
+
+/*
+ * preprocessors ought to look like this:
  */
 typedef struct preproc_ops {
     /*
@@ -273,7 +307,7 @@ typedef struct preproc_ops {
      * of the pass, an error reporting function, an evaluator
      * function, and a listing generator to talk to.
      */
-    void (*reset) (char *, int, efunc, evalfunc, ListGen *);
+    void (*reset) (char *, int, efunc, evalfunc, ListGen *, StrList **);
 
     /*
      * Called to fetch a line of preprocessed source. The line
@@ -348,7 +382,7 @@ enum {
  *
  * The bits are assigned as follows:
  *
- * Bits 0-7, 29: sizes
+ * Bits 0-7, 23, 29: sizes
  *  0:  8 bits (BYTE)
  *  1: 16 bits (WORD)
  *  2: 32 bits (DWORD)
@@ -357,6 +391,7 @@ enum {
  *  5: FAR
  *  6: NEAR
  *  7: SHORT
+ * 23: 256 bits (YWORD)
  * 29: 128 bits (OWORD)
  *
  * Bits 8-11 modifiers
@@ -371,17 +406,18 @@ enum {
  * 14: MEMORY (always has REGMEM attribute as well)
  * 15: REGMEM (valid EA operand)
  *
- * Bits 16-19: subclasses
+ * Bits 16-19, 28: subclasses
  * With REG_CDT:
  * 16: REG_CREG (CRx)
  * 17: REG_DREG (DRx)
  * 18: REG_TREG (TRx)
 
  * With REG_GPR:
- * 16: REG_ACCUM (AL, AX, EAX, RAX)
- * 17: REG_COUNT (CL, CX, ECX, RCX)
- * 18: REG_DATA  (DL, DX, EDX, RDX)
- * 19: REG_HIGH  (AH, CH, DH, BH)
+ * 16: REG_ACCUM  (AL, AX, EAX, RAX)
+ * 17: REG_COUNT  (CL, CX, ECX, RCX)
+ * 18: REG_DATA   (DL, DX, EDX, RDX)
+ * 19: REG_HIGH   (AH, CH, DH, BH)
+ * 28: REG_NOTACC (not REG_ACCUM)
  *
  * With REG_SREG:
  * 16: REG_CS
@@ -395,24 +431,29 @@ enum {
  * With XMMREG:
  * 16: XMM0
  *
+ * With YMMREG:
+ * 16: YMM0
+ *
  * With MEMORY:
  * 16: MEM_OFFS (this is a simple offset)
  * 17: IP_REL (IP-relative offset)
  *
  * With IMMEDIATE:
  * 16: UNITY (1)
- * 17: BYTENESS (-128..127)
+ * 17: BYTENESS16 (-128..127)
+ * 18: BYTENESS32 (-128..127)
+ * 19: BYTENESS64 (-128..127)
  *
- * Bits 20-26: register classes
+ * Bits 20-22, 24-27: register classes
  * 20: REG_CDT (CRx, DRx, TRx)
  * 21: RM_GPR (REG_GPR) (integer register)
  * 22: REG_SREG
- * 23: IP_REG (RIP or EIP) [unused]
  * 24: FPUREG
  * 25: RM_MMX (MMXREG)
  * 26: RM_XMM (XMMREG)
+ * 27: RM_YMM (YMMREG)
  *
- * Bits 27-29 & 31 are currently unallocated.
+ * Bit 31 is currently unallocated.
  *
  * 30: SAME_AS
  * Special flag only used in instruction patterns; means this operand
@@ -423,94 +464,102 @@ enum {
 typedef uint32_t opflags_t;
 
 /* Size, and other attributes, of the operand */
-#define BITS8		0x00000001L
-#define BITS16		0x00000002L
-#define BITS32		0x00000004L
-#define BITS64		0x00000008L   /* x64 and FPU only */
-#define BITS80		0x00000010L   /* FPU only */
-#define BITS128		0x20000000L
-#define FAR		0x00000020L   /* grotty: this means 16:16 or */
+#define BITS8		0x00000001U
+#define BITS16		0x00000002U
+#define BITS32		0x00000004U
+#define BITS64		0x00000008U   /* x64 and FPU only */
+#define BITS80		0x00000010U   /* FPU only */
+#define BITS128		0x20000000U
+#define BITS256		0x00800000U
+#define FAR		0x00000020U   /* grotty: this means 16:16 or */
                                        /* 16:32, like in CALL/JMP */
-#define NEAR		0x00000040L
-#define SHORT		0x00000080L   /* and this means what it says :) */
+#define NEAR		0x00000040U
+#define SHORT		0x00000080U   /* and this means what it says :) */
 
-#define SIZE_MASK	0x200000FFL   /* all the size attributes */
+#define SIZE_MASK	0x208000FFU   /* all the size attributes */
 
 /* Modifiers */
-#define MODIFIER_MASK	0x00000f00L
-#define TO		0x00000100L   /* reverse effect in FADD, FSUB &c */
-#define COLON		0x00000200L   /* operand is followed by a colon */
-#define STRICT		0x00000400L   /* do not optimize this operand */
+#define MODIFIER_MASK	0x00000f00U
+#define TO		0x00000100U   /* reverse effect in FADD, FSUB &c */
+#define COLON		0x00000200U   /* operand is followed by a colon */
+#define STRICT		0x00000400U   /* do not optimize this operand */
 
 /* Type of operand: memory reference, register, etc. */
-#define OPTYPE_MASK	0x0000f000L
-#define REGISTER	0x00001000L   /* register number in 'basereg' */
-#define IMMEDIATE	0x00002000L
-#define MEMORY		0x0000c000L
-#define REGMEM		0x00008000L   /* for r/m, ie EA, operands */
+#define OPTYPE_MASK	0x0000f000U
+#define REGISTER	0x00001000U   /* register number in 'basereg' */
+#define IMMEDIATE	0x00002000U
+#define MEMORY		0x0000c000U
+#define REGMEM		0x00008000U   /* for r/m, ie EA, operands */
 
 /* Register classes */
-#define REG_EA		0x00009000L   /* 'normal' reg, qualifies as EA */
-#define RM_GPR		0x00208000L   /* integer operand */
-#define REG_GPR		0x00209000L   /* integer register */
-#define REG8		0x00209001L   /*  8-bit GPR  */
-#define REG16		0x00209002L   /* 16-bit GPR */
-#define REG32		0x00209004L   /* 32-bit GPR */
-#define REG64		0x00209008L   /* 64-bit GPR */
-#define IP_REG		0x00801000L   /* RIP or EIP register */
-#define RIPREG		0x00801008L   /* RIP */
-#define EIPREG		0x00801004L   /* EIP */
-#define FPUREG		0x01001000L   /* floating point stack registers */
-#define FPU0		0x01011000L   /* FPU stack register zero */
-#define RM_MMX		0x02008000L   /* MMX operand */
-#define MMXREG		0x02009000L   /* MMX register */
-#define RM_XMM		0x04008000L   /* XMM (SSE) operand */
-#define XMMREG		0x04009000L   /* XMM (SSE) register */
-#define XMM0		0x04019000L   /* XMM register zero */
-#define REG_CDT		0x00101004L   /* CRn, DRn and TRn */
-#define REG_CREG	0x00111004L   /* CRn */
-#define REG_DREG	0x00121004L   /* DRn */
-#define REG_TREG	0x00141004L   /* TRn */
-#define REG_SREG	0x00401002L   /* any segment register */
-#define REG_CS		0x00411002L   /* CS */
-#define REG_DESS	0x00421002L   /* DS, ES, SS */
-#define REG_FSGS	0x00441002L   /* FS, GS */
-#define REG_SEG67	0x00481002L   /* Unimplemented segment registers */
+#define REG_EA		0x00009000U   /* 'normal' reg, qualifies as EA */
+#define RM_GPR		0x00208000U   /* integer operand */
+#define REG_GPR		0x00209000U   /* integer register */
+#define REG8		0x00209001U   /*  8-bit GPR  */
+#define REG16		0x00209002U   /* 16-bit GPR */
+#define REG32		0x00209004U   /* 32-bit GPR */
+#define REG64		0x00209008U   /* 64-bit GPR */
+#define FPUREG		0x01001000U   /* floating point stack registers */
+#define FPU0		0x01011000U   /* FPU stack register zero */
+#define RM_MMX		0x02008000U   /* MMX operand */
+#define MMXREG		0x02009000U   /* MMX register */
+#define RM_XMM		0x04008000U   /* XMM (SSE) operand */
+#define XMMREG		0x04009000U   /* XMM (SSE) register */
+#define XMM0		0x04019000U   /* XMM register zero */
+#define RM_YMM		0x08008000U   /* YMM (AVX) operand */
+#define YMMREG		0x08009000U   /* YMM (AVX) register */
+#define YMM0		0x08019000U   /* YMM register zero */
+#define REG_CDT		0x00101004U   /* CRn, DRn and TRn */
+#define REG_CREG	0x00111004U   /* CRn */
+#define REG_DREG	0x00121004U   /* DRn */
+#define REG_TREG	0x00141004U   /* TRn */
+#define REG_SREG	0x00401002U   /* any segment register */
+#define REG_CS		0x00411002U   /* CS */
+#define REG_DESS	0x00421002U   /* DS, ES, SS */
+#define REG_FSGS	0x00441002U   /* FS, GS */
+#define REG_SEG67	0x00481002U   /* Unimplemented segment registers */
 
-#define REG_RIP		0x00801008L   /* RIP relative addressing */
-#define REG_EIP		0x00801004L   /* EIP relative addressing */
+#define REG_RIP		0x00801008U   /* RIP relative addressing */
+#define REG_EIP		0x00801004U   /* EIP relative addressing */
 
 /* Special GPRs */
-#define REG_SMASK	0x000f0000L   /* a mask for the following */
-#define REG_ACCUM	0x00219000L   /* accumulator: AL, AX, EAX, RAX */
-#define REG_AL		0x00219001L
-#define REG_AX		0x00219002L
-#define REG_EAX		0x00219004L
-#define REG_RAX		0x00219008L
-#define REG_COUNT	0x00229000L   /* counter: CL, CX, ECX, RCX */
-#define REG_CL		0x00229001L
-#define REG_CX		0x00229002L
-#define REG_ECX		0x00229004L
-#define REG_RCX		0x00229008L
-#define REG_DL		0x00249001L   /* data: DL, DX, EDX, RDX */
-#define REG_DX		0x00249002L
-#define REG_EDX		0x00249004L
-#define REG_RDX		0x00249008L
-#define REG_HIGH	0x00289001L   /* high regs: AH, CH, DH, BH */
+#define REG_SMASK	0x100f0000U   /* a mask for the following */
+#define REG_ACCUM	0x00219000U   /* accumulator: AL, AX, EAX, RAX */
+#define REG_AL		0x00219001U
+#define REG_AX		0x00219002U
+#define REG_EAX		0x00219004U
+#define REG_RAX		0x00219008U
+#define REG_COUNT	0x10229000U   /* counter: CL, CX, ECX, RCX */
+#define REG_CL		0x10229001U
+#define REG_CX		0x10229002U
+#define REG_ECX		0x10229004U
+#define REG_RCX		0x10229008U
+#define REG_DL		0x10249001U   /* data: DL, DX, EDX, RDX */
+#define REG_DX		0x10249002U
+#define REG_EDX		0x10249004U
+#define REG_RDX		0x10249008U
+#define REG_HIGH	0x10289001U   /* high regs: AH, CH, DH, BH */
+#define REG_NOTACC	0x10000000U   /* non-accumulator register */
+#define REG8NA		0x10209001U   /*  8-bit non-acc GPR  */
+#define REG16NA		0x10209002U   /* 16-bit non-acc GPR */
+#define REG32NA		0x10209004U   /* 32-bit non-acc GPR */
+#define REG64NA		0x10209008U   /* 64-bit non-acc GPR */
 
 /* special types of EAs */
-#define MEM_OFFS	0x0001c000L   /* simple [address] offset - absolute! */
-#define IP_REL		0x0002c000L   /* IP-relative offset */
+#define MEM_OFFS	0x0001c000U   /* simple [address] offset - absolute! */
+#define IP_REL		0x0002c000U   /* IP-relative offset */
 
 /* memory which matches any type of r/m operand */
-#define MEMORY_ANY	(MEMORY|RM_GPR|RM_MMX|RM_XMM)
+#define MEMORY_ANY	(MEMORY|RM_GPR|RM_MMX|RM_XMM|RM_YMM)
 
 /* special type of immediate operand */
-#define UNITY		0x00012000L   /* for shift/rotate instructions */
-#define SBYTE		0x00022000L   /* for op r16/32,immediate instrs. */
+#define UNITY		0x00012000U   /* for shift/rotate instructions */
+#define SBYTE16		0x00022000U   /* for op r16,immediate instrs. */
+#define SBYTE32		0x00042000U   /* for op r32,immediate instrs. */
+#define SBYTE64		0x00082000U   /* for op r64,immediate instrs. */
 
 /* special flags */
-#define SAME_AS		0x40000000L
+#define SAME_AS		0x40000000U
 
 /* Register names automatically generated from regs.dat */
 #include "regs.h"
@@ -525,16 +574,17 @@ enum ccode {			/* condition code names */
 /*
  * REX flags
  */
-#define REX_OC		0x0200	/* DREX suffix has the OC0 bit set */
-#define REX_D		0x0100	/* Instruction uses DREX instead of REX */
-#define REX_H		0x80	/* High register present, REX forbidden */
-#define REX_P		0x40	/* REX prefix present/required */
-#define REX_L		0x20	/* Use LOCK prefix instead of REX.R */
-#define REX_W		0x08	/* 64-bit operand size */
-#define REX_R		0x04	/* ModRM reg extension */
-#define REX_X		0x02	/* SIB index extension */
-#define REX_B		0x01	/* ModRM r/m extension */
 #define REX_REAL	0x4f	/* Actual REX prefix bits */
+#define REX_B		0x01	/* ModRM r/m extension */
+#define REX_X		0x02	/* SIB index extension */
+#define REX_R		0x04	/* ModRM reg extension */
+#define REX_W		0x08	/* 64-bit operand size */
+#define REX_L		0x20	/* Use LOCK prefix instead of REX.R */
+#define REX_P		0x40	/* REX prefix present/required */
+#define REX_H		0x80	/* High register present, REX forbidden */
+#define REX_D		0x0100	/* Instruction uses DREX instead of REX */
+#define REX_OC		0x0200	/* DREX suffix has the OC0 bit set */
+#define REX_V		0x0400	/* Instruction uses VEX instead of REX */
 
 /*
  * Note that because segment registers may be used as instruction
@@ -612,7 +662,8 @@ enum prefix_pos {
     MAXPREFIX			/* Total number of prefix slots */
 };
 
-#define MAX_OPERANDS 4
+/* If you need to change this, also change it in insns.pl */
+#define MAX_OPERANDS 5
 
 typedef struct insn {		/* an instruction itself */
     char *label;		/* the label defined, or NULL */
@@ -628,7 +679,9 @@ typedef struct insn {		/* an instruction itself */
     int32_t times;              /* repeat count (TIMES prefix) */
     int forw_ref;               /* is there a forward reference? */
     int rex;			/* Special REX Prefix */
-    int drexdst;		/* Destination register for DREX suffix */
+    int drexdst;		/* Destination register for DREX/VEX suffix */
+    int vex_m;			/* M register for VEX prefix */
+    int vex_wlp;		/* W, P and L information for VEX prefix */
 } insn;
 
 enum geninfo { GI_SWITCH };
@@ -709,7 +762,8 @@ struct ofmt {
      * The `type' argument specifies the type of output data, and
      * usually the size as well: its contents are described below.
      */
-    void (*output) (int32_t segto, const void *data, uint32_t type,
+    void (*output) (int32_t segto, const void *data,
+		    enum out_type type, uint64_t size,
                     int32_t segment, int32_t wrt);
 
     /*
@@ -740,7 +794,7 @@ struct ofmt {
      * been an EXTERN, a COMMON or a GLOBAL. The distinction should
      * be obvious to the output format from the other parameters.
      */
-    void (*symdef) (char *name, int32_t segment, int32_t offset,
+    void (*symdef) (char *name, int32_t segment, int64_t offset,
 		    int is_global, char *special);
 
     /*
@@ -822,31 +876,6 @@ struct ofmt {
     void (*cleanup) (int debuginfo);
 };
 
-/*
- * values for the `type' parameter to an output function. Each one
- * must have the actual number of _bytes_ added to it.
- *
- * Exceptions are OUT_RELxADR, which denote an x-byte relocation
- * which will be a relative jump. For this we need to know the
- * distance in bytes from the start of the relocated record until
- * the end of the containing instruction. _This_ is what is stored
- * in the size part of the parameter, in this case.
- *
- * Also OUT_RESERVE denotes reservation of N bytes of BSS space,
- * and the contents of the "data" parameter is irrelevant.
- *
- * The "data" parameter for the output function points to a "int32_t",
- * containing the address in question, unless the type is
- * OUT_RAWDATA, in which case it points to an "uint8_t"
- * array.
- */
-#define OUT_RAWDATA 0x00000000UL
-#define OUT_ADDRESS 0x10000000UL
-#define OUT_REL2ADR 0x20000000UL
-#define OUT_REL4ADR 0x30000000UL
-#define OUT_RESERVE 0x40000000UL
-#define OUT_TYPMASK 0xF0000000UL
-#define OUT_SIZMASK 0x0FFFFFFFUL
 
 /*
  * ------------------------------------------------------------
@@ -887,7 +916,7 @@ struct dfmt {
      * would be called before the output format version.
      */
 
-    void (*debug_deflabel) (char *name, int32_t segment, int32_t offset,
+    void (*debug_deflabel) (char *name, int32_t segment, int64_t offset,
                             int is_global, char *special);
     /*
      * debug_directive - called whenever a DEBUG directive other than 'LINE'
@@ -936,6 +965,7 @@ struct dfmt {
 #define TY_QWORD   0x30
 #define TY_TBYTE   0x38
 #define TY_OWORD   0x40
+#define TY_YWORD   0x48
 #define TY_COMMON  0xE0
 #define TY_SEG     0xE8
 #define TY_EXTERN  0xF0
@@ -956,23 +986,10 @@ enum special_tokens {
     SPECIAL_ENUM_START = PREFIX_ENUM_LIMIT,
     S_ABS = SPECIAL_ENUM_START,
     S_BYTE, S_DWORD, S_FAR, S_LONG, S_NEAR, S_NOSPLIT,
-    S_OWORD, S_QWORD, S_REL, S_SHORT, S_STRICT, S_TO, S_TWORD, S_WORD,
+    S_OWORD, S_QWORD, S_REL, S_SHORT, S_STRICT, S_TO, S_TWORD, S_WORD, S_YWORD,
     SPECIAL_ENUM_LIMIT
 };
 
-/*
- * -----
- * Other
- * -----
- */
-
-/*
- * This is a useful #define which I keep meaning to use more often:
- * the number of elements of a statically defined array.
- */
-
-#define elements(x)     ( sizeof(x) / sizeof(*(x)) )
-
 /*
  * -----
  * Global modes
@@ -988,6 +1005,7 @@ enum special_tokens {
  */
 
 extern int pass0;
+extern int passn;		/* Actual pass number */
 
 extern bool tasm_compatible_mode;
 extern int optimizing;
