@@ -3,7 +3,7 @@
  *
  * The Netwide Assembler is copyright (C) 1996 Simon Tatham and
  * Julian Hall. All rights reserved. The software is
- * redistributable under the licence given in the file "Licence"
+ * redistributable under the license given in the file "LICENSE"
  * distributed in the NASM archive.
  */
 
@@ -68,6 +68,9 @@
 /* Flag which version of COFF we are currently outputting. */
 static bool win32, win64;
 
+static int32_t imagebase_sect;
+#define WRT_IMAGEBASE "..imagebase"
+
 struct Reloc {
     struct Reloc *next;
     int32_t address;		/* relative to _start_ of section */
@@ -77,10 +80,18 @@ struct Reloc {
         ABS_SYMBOL,
         REAL_SYMBOLS
     } symbase;                  /* relocation for symbol number :) */
-    bool relative;
-    bool size64;
+    int16_t type;
 };
 
+/* possible values for Reloc->type */
+#define IMAGE_REL_AMD64_ADDR64		0x0001
+#define IMAGE_REL_AMD64_ADDR32		0x0002
+#define IMAGE_REL_AMD64_ADDR32NB	0x0003
+#define IMAGE_REL_AMD64_REL32		0x0004
+#define IMAGE_REL_I386_DIR32		0x0006
+#define IMAGE_REL_I386_DIR32NB		0x0007
+#define IMAGE_REL_I386_REL32		0x0014
+
 struct Symbol {
     char name[9];
     int32_t strpos;		/* string table position of name */
@@ -88,6 +99,8 @@ struct Symbol {
     int section;                /* section number where it's defined
                                  * - in COFF codes, not NASM codes */
     bool is_global;              /* is it a global symbol or not? */
+    int16_t type;		/* 0 - notype, 0x20 - function */
+    int32_t namlen;		/* full name length */
 };
 
 static FILE *coffp;
@@ -147,11 +160,15 @@ static void coff_win32_init(FILE * fp, efunc errfunc,
 static void coff_win64_init(FILE * fp, efunc errfunc,
                             ldfunc ldef, evalfunc eval)
 {
+    extern struct ofmt of_win64;
+
     maxbits = 64;
     win32 = false; win64 = true;
     (void)ldef;                 /* placate optimizers */
     (void)eval;
     coff_gen_init(fp, errfunc);
+    imagebase_sect = seg_alloc()+1;
+    ldef(WRT_IMAGEBASE,imagebase_sect,0,NULL,false,false,&of_win64,errfunc);
 }
 
 static void coff_std_init(FILE * fp, efunc errfunc, ldfunc ldef,
@@ -341,6 +358,10 @@ static int32_t coff_section_names(char *name, int pass, int *bits)
                 flags = RDATA_FLAGS;
             else if (!strcmp(name, ".bss"))
                 flags = BSS_FLAGS;
+	    else if (win64 && !strcmp(name, ".pdata"))
+		flags = 0x40300040; /* rdata align=4 */
+	    else if (win64 && !strcmp(name, ".xdata"))
+		flags = 0x40400040; /* rdate align=8 */
             else
                 flags = TEXT_FLAGS;
         }
@@ -358,7 +379,7 @@ static int32_t coff_section_names(char *name, int pass, int *bits)
     return sects[i]->index;
 }
 
-static void coff_deflabel(char *name, int32_t segment, int32_t offset,
+static void coff_deflabel(char *name, int32_t segment, int64_t offset,
                           int is_global, char *special)
 {
     int pos = strslen + 4;
@@ -369,7 +390,8 @@ static void coff_deflabel(char *name, int32_t segment, int32_t offset,
               " special symbol types");
 
     if (name[0] == '.' && name[1] == '.' && name[2] != '@') {
-        error(ERR_NONFATAL, "unrecognized special symbol `%s'", name);
+	if (strcmp(name,WRT_IMAGEBASE))
+            error(ERR_NONFATAL, "unrecognized special symbol `%s'", name);
         return;
     }
 
@@ -382,6 +404,7 @@ static void coff_deflabel(char *name, int32_t segment, int32_t offset,
     sym = saa_wstruct(syms);
 
     sym->strpos = pos;
+    sym->namlen = strlen(name);
     if (pos == -1)
         strcpy(sym->name, name);
     sym->is_global = !!is_global;
@@ -418,12 +441,10 @@ static void coff_deflabel(char *name, int32_t segment, int32_t offset,
 }
 
 static int32_t coff_add_reloc(struct Section *sect, int32_t segment,
-                           int relative, int size64)
+				int16_t type)
 {
     struct Reloc *r;
 
-    (void)size64;
-
     r = *sect->tail = nasm_malloc(sizeof(struct Reloc));
     sect->tail = &r->next;
     r->next = NULL;
@@ -443,7 +464,7 @@ static int32_t coff_add_reloc(struct Section *sect, int32_t segment,
         if (r->symbase == REAL_SYMBOLS)
             r->symbol = raa_read(bsym, segment);
     }
-    r->relative = relative;
+    r->type = type;
 
     sect->nrelocs++;
 
@@ -456,21 +477,19 @@ static int32_t coff_add_reloc(struct Section *sect, int32_t segment,
         return 0;
 }
 
-static void coff_out(int32_t segto, const void *data, uint32_t type,
+static void coff_out(int32_t segto, const void *data,
+		     enum out_type type, uint64_t size,
                      int32_t segment, int32_t wrt)
 {
     struct Section *s;
-    int32_t realbytes = type & OUT_SIZMASK;
     uint8_t mydata[8], *p;
     int i;
 
-    if (wrt != NO_SEG) {
+    if (wrt != NO_SEG && !win64) {
         wrt = NO_SEG;           /* continue to do _something_ */
         error(ERR_NONFATAL, "WRT not supported by COFF output formats");
     }
 
-    type &= OUT_TYPMASK;
-
     /*
      * handle absolute-assembly (structure definitions)
      */
@@ -495,14 +514,19 @@ static void coff_out(int32_t segto, const void *data, uint32_t type,
             s = sects[nsects - 1];
     }
 
+    /* magically default to 'wrt ..imagebase' in .pdata and .xdata */
+    if (win64 && wrt == NO_SEG &&
+	(!strcmp(s->name,".pdata") || !strcmp(s->name,".xdata")))
+	wrt = imagebase_sect;
+
     if (!s->data && type != OUT_RESERVE) {
         error(ERR_WARNING, "attempt to initialize memory in"
               " BSS section `%s': ignored", s->name);
         if (type == OUT_REL2ADR)
-            realbytes = 2;
+            size = 2;
         else if (type == OUT_REL4ADR)
-            realbytes = 4;
-        s->len += realbytes;
+            size = 4;
+        s->len += size;
         return;
     }
 
@@ -510,16 +534,16 @@ static void coff_out(int32_t segto, const void *data, uint32_t type,
         if (s->data) {
             error(ERR_WARNING, "uninitialised space declared in"
                   " non-BSS section `%s': zeroing", s->name);
-            coff_sect_write(s, NULL, realbytes);
+            coff_sect_write(s, NULL, size);
         } else
-            s->len += realbytes;
+            s->len += size;
     } else if (type == OUT_RAWDATA) {
         if (segment != NO_SEG)
             error(ERR_PANIC, "OUT_RAWDATA with other than NO_SEG");
-        coff_sect_write(s, data, realbytes);
+        coff_sect_write(s, data, size);
     } else if (type == OUT_ADDRESS) {
         if (!(win64)) {
-            if (realbytes != 4 && (segment != NO_SEG || wrt != NO_SEG))
+            if (size != 4 && (segment != NO_SEG || wrt != NO_SEG))
                 error(ERR_NONFATAL, "COFF format does not support non-32-bit"
                       " relocations");
             else {
@@ -532,33 +556,29 @@ static void coff_out(int32_t segto, const void *data, uint32_t type,
                         error(ERR_NONFATAL, "COFF format does not support"
                               " segment base references");
                     } else
-                        fix = coff_add_reloc(s, segment, false, false);
+                        fix = coff_add_reloc(s, segment, IMAGE_REL_I386_DIR32);
                 }
                 p = mydata;
-                WRITELONG(p, *(int32_t *)data + fix);
-                coff_sect_write(s, mydata, realbytes);
+                WRITELONG(p, *(int64_t *)data + fix);
+                coff_sect_write(s, mydata, size);
             }
         } else {
             int32_t fix = 0;
             p = mydata;
-            if (realbytes == 8) {
-/*            if (segment != NO_SEG || wrt != NO_SEG) {
-                if (wrt != NO_SEG) {
-                    error(ERR_NONFATAL, "COFF format does not support"
-                          " WRT types");
-                } else if (segment % 2) {
-                    error(ERR_NONFATAL, "COFF format does not support"
-                          " segment base references");
-                } else
-                    fix = coff_add_reloc(s, segment, false);
-            } */
-                fix = coff_add_reloc(s, segment, false, true);
+            if (size == 8) {
+	        if (wrt == imagebase_sect) {
+		    error(ERR_NONFATAL, "operand size mismatch: 'wrt "
+			WRT_IMAGEBASE "' is a 32-bit operand");
+		}
+                fix = coff_add_reloc(s, segment, IMAGE_REL_AMD64_ADDR64);
                 WRITEDLONG(p, *(int64_t *)data + fix);
-                coff_sect_write(s, mydata, realbytes);
+                coff_sect_write(s, mydata, size);
             } else {
-                fix = coff_add_reloc(s, segment, false, false);
-                WRITELONG(p, *(int32_t *)data + fix);
-                coff_sect_write(s, mydata, realbytes);
+                fix = coff_add_reloc(s, segment,
+			wrt == imagebase_sect ?	IMAGE_REL_AMD64_ADDR32NB:
+						IMAGE_REL_AMD64_ADDR32);
+                WRITELONG(p, *(int64_t *)data + fix);
+                coff_sect_write(s, mydata, size);
             }
         }
     } else if (type == OUT_REL2ADR) {
@@ -576,12 +596,13 @@ static void coff_out(int32_t segto, const void *data, uint32_t type,
                 error(ERR_NONFATAL, "COFF format does not support"
                       " segment base references");
             } else
-                fix = coff_add_reloc(s, segment, true, false);
+                fix = coff_add_reloc(s, segment,
+			win64 ? IMAGE_REL_AMD64_REL32 : IMAGE_REL_I386_REL32);
             p = mydata;
             if (win32 | win64) {
-                WRITELONG(p, *(int32_t *)data + 4 - realbytes + fix);
+                WRITELONG(p, *(int64_t *)data + 4 - size + fix);
             } else {
-                WRITELONG(p, *(int32_t *)data - (realbytes + s->len) + fix);
+                WRITELONG(p, *(int64_t *)data - (size + s->len) + fix);
             }
             coff_sect_write(s, mydata, 4L);
         }
@@ -691,6 +712,54 @@ static int coff_directives(char *directive, char *value, int pass)
         }
         AddExport(name);
         return 1;
+    } else if (win32 && !strcmp(directive,"safeseh")) {
+	static int sxseg=-1;
+	int i;
+	if (sxseg==-1)
+	{   for (i = 0; i < nsects; i++)
+		if (!strcmp(".sxdata",sects[i]->name))
+		    break;
+	    if (i == nsects)
+		sxseg = coff_make_section(".sxdata",0x200);
+	    else
+		sxseg = i;
+	}
+	if (pass==2) {
+	    uint32_t n;
+	    saa_rewind(syms);
+	    for (n = 0; n < nsyms; n++) {
+		struct Symbol *sym = saa_rstruct(syms);
+		bool equals;
+
+		/* sym->strpos is biased by 4, because symbol
+		 * table is prefixed with table length */
+		if (sym->strpos >=4) {
+		    char *name = nasm_malloc(sym->namlen+1);
+		    saa_fread(strs,sym->strpos-4,name,sym->namlen);
+		    name[sym->namlen]='\0';
+		    equals = !strcmp(value,name);
+		    nasm_free(name);
+		}
+		else
+		    equals = !strcmp(value,sym->name);
+
+		if (equals) {
+		    /* this value arithmetics effectively reflects
+		     * initsym in coff_write(): 2 for file, 1 for
+		     * .absolute and two per each section */
+		    unsigned char value[4],*p=value;
+		    WRITELONG(p,n + 2 + 1 + nsects*2);
+		    coff_sect_write(sects[sxseg],value,4);
+		    sym->type = 0x20;
+		    break;
+		}
+	    }
+	    if (n == nsyms) {
+		error(ERR_NONFATAL,
+		      "`safeseh' directive requires valid symbol");
+	    }
+	}
+	return 1;
     }
     return 0;
 }
@@ -701,6 +770,21 @@ static void coff_write(void)
     int i;
 
     BuildExportTable();         /* fill in the .drectve section with -export's */
+
+    if (win32) {
+	/* add default value for @feat.00, this allows to 'link /safeseh' */
+	uint32_t n;
+
+	saa_rewind(syms);
+	for (n = 0; n < nsyms; n++) {
+            struct Symbol *sym = saa_rstruct(syms);
+	    if (sym->strpos == -1 && !strcmp("@feat.00",sym->name))
+		break;
+	}
+	if (n == nsyms)
+	    coff_deflabel("@feat.00",NO_SEG,1,0,NULL);
+    }
+
     /*
      * Work out how big the file will get. Calculate the start of
      * the `real' symbols at the same time.
@@ -794,21 +878,12 @@ static void coff_write_relocs(struct Section *s)
                                 r->symbase == ABS_SYMBOL ? initsym - 1 :
                                 r->symbase == SECT_SYMBOLS ? 2 : 0),
                    coffp);
-        /*
-         * Strange: Microsoft's COFF documentation says 0x03 for an
-         * absolute relocation, but both Visual C++ and DJGPP agree
-         * that in fact it's 0x06. I'll use 0x06 until someone
-         * argues. ***** UPDATE: PE/COFF Ver.8 docs confirm this -kkanios *****
-         */
-        if (win64)
-            fwriteint16_t(r->relative ? 0x04 : r->size64 ? 0x01 : 0x02, coffp);
-        else
-            fwriteint16_t(r->relative ? 0x14 : 0x06, coffp);
+	fwriteint16_t(r->type, coffp);
     }
 }
 
 static void coff_symbol(char *name, int32_t strpos, int32_t value,
-                        int section, int type, int aux)
+                        int section, int type, int storageclass, int aux)
 {
     char padname[8];
 
@@ -822,8 +897,8 @@ static void coff_symbol(char *name, int32_t strpos, int32_t value,
     }
     fwriteint32_t(value, coffp);
     fwriteint16_t(section, coffp);
-    fwriteint16_t(0, coffp);
-    fputc(type, coffp);
+    fwriteint16_t(type, coffp);
+    fputc(storageclass, coffp);
     fputc(aux, coffp);
 }
 
@@ -835,7 +910,7 @@ static void coff_write_symbols(void)
     /*
      * The `.file' record, and the file name auxiliary record.
      */
-    coff_symbol(".file", 0L, 0L, -2, 0x67, 1);
+    coff_symbol(".file", 0L, 0L, -2, 0, 0x67, 1);
     memset(filename, 0, 18);
     strncpy(filename, coff_infile, 18);
     fwrite(filename, 18, 1, coffp);
@@ -845,8 +920,8 @@ static void coff_write_symbols(void)
      */
     memset(filename, 0, 18);    /* useful zeroed buffer */
 
-    for (i = 0; i < nsects; i++) {
-        coff_symbol(sects[i]->name, 0L, 0L, i + 1, 3, 1);
+    for (i = 0; i < (uint32_t) nsects; i++) {
+        coff_symbol(sects[i]->name, 0L, 0L, i + 1, 0, 3, 1);
         fwriteint32_t(sects[i]->len, coffp);
         fwriteint16_t(sects[i]->nrelocs, coffp);
         fwrite(filename, 12, 1, coffp);
@@ -855,7 +930,7 @@ static void coff_write_symbols(void)
     /*
      * The absolute symbol, for relative-to-absolute relocations.
      */
-    coff_symbol(".absolut", 0L, 0L, -1, 3, 0);
+    coff_symbol(".absolut", 0L, 0L, -1, 0, 3, 0);
 
     /*
      * The real symbols.
@@ -865,7 +940,7 @@ static void coff_write_symbols(void)
         struct Symbol *sym = saa_rstruct(syms);
         coff_symbol(sym->strpos == -1 ? sym->name : NULL,
                     sym->strpos, sym->value, sym->section,
-                    sym->is_global ? 2 : 3, 0);
+                    sym->type, sym->is_global ? 2 : 3, 0);
     }
 }
 
@@ -893,6 +968,9 @@ static const char *coff_stdmac[] = {
     "%imacro export 1+.nolist",
     "[export %1]",
     "%endmacro",
+    "%imacro safeseh 1.nolist",
+    "[safeseh %1]",
+    "%endmacro",
     NULL
 };
 
