@@ -3,7 +3,7 @@
  *
  * The Netwide Assembler is copyright (C) 1996 Simon Tatham and
  * Julian Hall. All rights reserved. The software is
- * redistributable under the licence given in the file "Licence"
+ * redistributable under the license given in the file "LICENSE"
  * distributed in the NASM archive.
  */
 
@@ -386,16 +386,14 @@ static void add_reloc(struct section *sect, int32_t section,
     ++sect->nreloc;
 }
 
-static void macho_output(int32_t secto, const void *data, uint32_t type,
+static void macho_output(int32_t secto, const void *data,
+			 enum out_type type, uint64_t size,
                          int32_t section, int32_t wrt)
 {
     struct section *s, *sbss;
-    int32_t realbytes = type & OUT_SIZMASK;
     int32_t addr;
     uint8_t mydata[4], *p;
 
-    type &= OUT_TYPMASK;
-
     if (wrt != NO_SEG) {
         wrt = NO_SEG;
         error(ERR_NONFATAL, "WRT not supported by Mach-O output format");
@@ -430,18 +428,18 @@ static void macho_output(int32_t secto, const void *data, uint32_t type,
 
         switch (type) {
         case OUT_REL2ADR:
-            realbytes = 2;
+            size = 2;
             break;
 
         case OUT_REL4ADR:
-            realbytes = 4;
+            size = 4;
             break;
 
         default:
             break;
         }
 
-        s->size += realbytes;
+        s->size += size;
         return;
     }
 
@@ -452,9 +450,9 @@ static void macho_output(int32_t secto, const void *data, uint32_t type,
                   " %s section: zeroing",
                   get_section_name_by_index(secto));
 
-            sect_write(s, NULL, realbytes);
+            sect_write(s, NULL, size);
         } else
-            s->size += realbytes;
+            s->size += size;
 
         break;
 
@@ -462,28 +460,23 @@ static void macho_output(int32_t secto, const void *data, uint32_t type,
         if (section != NO_SEG)
             error(ERR_PANIC, "OUT_RAWDATA with other than NO_SEG");
 
-        sect_write(s, data, realbytes);
+        sect_write(s, data, size);
         break;
 
     case OUT_ADDRESS:
-        addr = *(int32_t *)data;
+        addr = *(int64_t *)data;
 
         if (section != NO_SEG) {
             if (section % 2) {
                 error(ERR_NONFATAL, "Mach-O format does not support"
                       " section base references");
             } else
-                add_reloc(s, section, 0, realbytes);
+                add_reloc(s, section, 0, size);
         }
 
         p = mydata;
-
-        if (realbytes == 2)
-            WRITESHORT(p, addr);
-        else
-            WRITELONG(p, addr);
-
-        sect_write(s, mydata, realbytes);
+	WRITEADDR(p, addr, size);
+        sect_write(s, mydata, size);
         break;
 
     case OUT_REL2ADR:
@@ -497,7 +490,7 @@ static void macho_output(int32_t secto, const void *data, uint32_t type,
             add_reloc(s, section, 1, 2);
 
         p = mydata;
-        WRITESHORT(p, *(int32_t *)data - (realbytes + s->size));
+        WRITESHORT(p, *(int32_t *)data - (size + s->size));
         sect_write(s, mydata, 2L);
         break;
 
@@ -512,7 +505,7 @@ static void macho_output(int32_t secto, const void *data, uint32_t type,
             add_reloc(s, section, 1, 4);
 
         p = mydata;
-        WRITELONG(p, *(int32_t *)data - (realbytes + s->size));
+        WRITELONG(p, *(int32_t *)data - (size + s->size));
         sect_write(s, mydata, 4L);
         break;
 
@@ -559,7 +552,7 @@ static int32_t macho_section(char *name, int pass, int *bits)
                 s->data = saa_init(1L);
                 s->index = seg_alloc();
                 s->relocs = NULL;
-                s->align = DEFAULT_SECTION_ALIGNMENT;
+                s->align = -1;
 
                 xstrncpy(s->segname, sm->segname);
                 xstrncpy(s->sectname, sm->sectname);
@@ -599,7 +592,8 @@ static int32_t macho_section(char *name, int pass, int *bits)
                         }
 
                         if ((-1 != originalIndex)
-                            && (s->align != newAlignment)) {
+                            && (s->align != newAlignment)
+                           && (s->align != -1)) {
                             error(ERR_PANIC,
                                   "section \"%s\" has already been specified "
                                       "with alignment %d, conflicts with new "
@@ -631,7 +625,7 @@ static int32_t macho_section(char *name, int pass, int *bits)
     return NO_SEG;
 }
 
-static void macho_symdef(char *name, int32_t section, int32_t offset,
+static void macho_symdef(char *name, int32_t section, int64_t offset,
                          int is_global, char *special)
 {
     struct symbol *sym;
@@ -912,6 +906,8 @@ static uint32_t macho_write_segment (uint32_t offset)
 	    fwriteint32_t(offset, machofp);
 	    /* Write out section alignment, as a power of two.
 	       e.g. 32-bit word alignment would be 2 (2^^2 = 4).  */
+	    if (s->align == -1)
+		s->align = DEFAULT_SECTION_ALIGNMENT;
 	    fwriteint32_t(s->align, machofp);
 	    /* To be compatible with cctools as we emit
 	       a zero reloff if we have no relocations.  */
@@ -1001,11 +997,10 @@ static void macho_write_section (void)
 	       offset which we already have. The linker takes care
 	       of the rest of the address.  */
 	    if (!r->ext) {
-		/* add sizes of previous sections to current offset */
-		for (s2 = sects, fi = 1;
-		     s2 != NULL && fi < r->snum; s2 = s2->next, fi++)
-		    if ((s2->flags & SECTION_TYPE) != S_ZEROFILL)
-			l += s2->size;
+           /* add sizes of previous sections to current offset */
+           for (s2 = sects, fi = 1;
+                s2 != NULL && fi < r->snum; s2 = s2->next, fi++)
+               l += s2->size;
 	    }
 
 	    /* write new offset back */
@@ -1111,6 +1106,7 @@ static void macho_fixup_relocs (struct reloc *r)
 		sym = undefsyms[i];
 		if (sym->initial_snum == r->snum) {
 		    r->snum = sym->snum;
+                   break;
 		}
 	    }
 	}
@@ -1280,14 +1276,14 @@ static void macho_cleanup(int debuginfo)
     raa_free(extsyms);
 
     if (syms) {
-	while (syms->next) {
-	    sym = syms;
-	    syms = syms->next;
+    while (syms->next) {
+       sym = syms;
+       syms = syms->next;
 
-	    nasm_free (sym);
-	}
+       nasm_free (sym);
     }
 }
+}
 
 /* Debugging routines.  */
 static void debug_reloc (struct reloc *r)
